from PyQt5 import QtGui, QtWidgets
from PyQt5.QtCore import pyqtSignal, QObject
import sys
import tables
from collections import OrderedDict

import numpy as np
import datetime
from pathlib import Path
from ctypes import CFUNCTYPE
if 'win32' in sys.platform:
    from ctypes import WINFUNCTYPE

from pyqtgraph import dockarea
import enum
import os
import re
import importlib
import inspect
from numba import jit

plot_colors = ['r', 'g','b',  'c', 'm', 'y', 'k',' w']
Cb = 1.602176e-19  # coulomb
h = 6.626068e-34  # J.s
c = 2.997924586e8  # m.s-1

def Enm2cmrel(E_nm, ref_wavelength=515):
    return 1/(ref_wavelength*1e-7)-1/(E_nm*1e-7)

def Ecmrel2Enm(Ecmrel, ref_wavelength=515):
    Ecm = Ecmrel+1/(ref_wavelength*1e-7)
    return 1/(Ecm*1e-7)


def eV2nm(E_eV):
    E_J = E_eV * Cb
    E_freq = E_J / h
    E_nm = c / E_freq * 1e9
    return E_nm


def nm2eV(E_nm):
    E_freq = c / E_nm * 1e9;
    E_J = E_freq * h;
    E_eV = E_J / Cb;
    return E_eV


def E_J2eV(E_J):
    E_eV = E_J / Cb;
    return E_eV


def eV2cm(E_eV):
    E_nm = eV2nm(E_eV)
    E_cm = 1 / (E_nm * 1e-7);
    return E_cm

def nm2cm(E_nm):
    return 1/(E_nm*1e7)

def cm2nm(E_cm):
    return 1 / (E_cm * 1e-7)


def eV2E_J(E_eV):
    E_J = E_eV * Cb;
    return E_J


def eV2radfs(E_eV):
    E_J = E_eV * Cb
    E_freq = E_J / h
    E_radfs = E_freq * 2 * np.pi / 1e15
    return E_radfs


def l2w(x, speedlight=300):
    # y=l2w(x,c)
    # speedlight is the speed of light =300 nm/fs
    y = 2 * np.pi * speedlight / x
    return y


def capitalize(string):
    """
    Returns same string but with first letter capitalized
    Parameters
    ----------
    string: (str)

    Returns
    -------
    str
    """
    return string.capitalize()[0]+string[1:]

def uncapitalize(string):
    return string.lower()[0] + string[1:]


class ListPicker(QObject):

    def __init__(self,list_str):
        super(ListPicker,self).__init__()
        self.list = list_str

    def pick_dialog(self):
        self.dialog = QtWidgets.QDialog()
        self.dialog.setMinimumWidth(500)
        vlayout = QtWidgets.QVBoxLayout()


        self.list_widget = QtWidgets.QListWidget()
        self.list_widget.addItems(self.list)

        vlayout.addWidget(self.list_widget, 10)
        self.dialog.setLayout(vlayout)

        buttonBox = QtWidgets.QDialogButtonBox();
        buttonBox.addButton('Apply', buttonBox.AcceptRole)
        buttonBox.accepted.connect(self.dialog.accept)
        buttonBox.addButton('Cancel', buttonBox.RejectRole)
        buttonBox.rejected.connect(self.dialog.reject)

        vlayout.addWidget(buttonBox)
        self.dialog.setWindowTitle('Select an entry in the list')

        res = self.dialog.show()

        pass
        if res == self.dialog.Accepted:
            # save preset parameters in a xml file
            return  [self.list_widget.currentIndex(), self.list_widget.currentItem().text()]
        else:
            return [-1, ""]

def get_data_dimension(arr, scan_type='scan1D', remove_scan_dimension=False):
    dimension = len(arr.shape)
    if dimension == 1:
        if arr.size == 1:
            dimension = 0
    if remove_scan_dimension:
        if scan_type.lower() == 'scan1d':
            dimension -= 1
        elif scan_type.lower() == 'scan2d':
            dimension -= 2

    return arr.shape, '{:d}D'.format(dimension), arr.size

def scroll_log(scroll_val, min_val , max_val):
    """
    Convert a scroll value [0-100] to a log scale between min_val and max_val
    Parameters
    ----------
    scroll
    min_val
    max_val
    Returns
    -------

    """
    scr = scroll_val
    value = scr * (np.log10(max_val)-np.log10(min_val))/100+ np.log10(min_val)
    return 10**value

def scroll_linear(scroll_val, min_val , max_val):
    """
    Convert a scroll value [0-100] to a linear scale between min_val and max_val
    Parameters
    ----------
    scroll
    min_val
    max_val
    Returns
    -------

    """
    scr = scroll_val
    value = scr * (max_val-min_val)/100+ min_val
    return value

def extract_TTTR_histo_every_pixels(nanotimes, markers, marker = 65, Nx = 1, Ny = 1, Ntime = 512, ind_line_offset = 0,
                                    channel = 0):
    """
    Extract histograms from photon tags and attributes them in the given pixel of the FLIM
    The marker is used to check where a new line within the image starts
    Parameters
    ----------
    nanotimes: (ndarray of uint16) photon arrival times (in timeharp units)
    markers: (ndarray of uint8) markers: 0 means the corresponding nanotime is a photon on detector 0,
                                         1 means the corresponding nanotime is a photon on detector 1,
                                         65 => Marker 1 event
                                         66 => Marker 2 event
                                         ...
                                         79 => Marker 15 event
                                         127 =>overflow
    marker: (int) the marker value corresponding to a new Y line within the image (for instance 65)
    Nx: (int) the number of pixels along the xaxis
    Ny: (int) the number of pixels along the yaxis
    Ntime: (int) the number of pixels alond the time axis
    ind_line_offset: (int) the offset of previously read lines
    channel: (int) marker of the specific channel (0 or 1) for channel 1 or 2

    Returns
    -------
    ndarray: FLIM hypertemporal image in the order (X, Y, time)
    """
    bins = np.linspace(0, Ntime, Ntime + 1, dtype=np.uint32)
    nanotimes = nanotimes[np.logical_or(markers == marker, markers == channel)]
    markers = markers[np.logical_or(markers == marker, markers == channel)]
    indexes_new_line = np.squeeze(np.argwhere(markers == marker)).astype(np.uint64)
    datas = np.zeros((Nx, Ny, Ntime), dtype=np.int64)

    if indexes_new_line.size == 0:
        indexes_new_line = np.array([0, nanotimes.size], dtype=np.uint64)
    # print(indexes_new_line)
    for ind_line in range(indexes_new_line.size - 1):
        # print(ind_line)
        data_line_tmp = nanotimes[int(indexes_new_line[ind_line] + 1):indexes_new_line[ind_line + 1]]
        ix = np.int((ind_line + ind_line_offset) % Nx)
        iy = np.int(((ind_line + ind_line_offset) // Nx) % Ny)
        datas[ix, iy, :] += np.histogram(data_line_tmp, bins, range=None)[0]

    return datas


def getLineInfo():
    return "in {:s}, method: {:s}, line: {:d}: ".format(os.path.split(inspect.stack()[1][1])[1], inspect.stack()[1][3], inspect.stack()[1][2])

class ScanParameters(object):
    def __init__(self, Nsteps=0,axis_1_indexes=[],axis_2_indexes=[],axis_1_unique=[],axis_2_unique=[],
                 positions=[]):
        super(ScanParameters, self).__init__()
        self.positions = positions
        self.axis_2D_1 = axis_1_unique
        self.axis_2D_2 = axis_2_unique
        self.axis_2D_1_indexes = axis_1_indexes
        self.axis_2D_2_indexes = axis_2_indexes
        self.Nsteps = Nsteps

    def __repr__(self):
        return 'Scanner with {:d} positions and shape:({:d}, {:d})'.format(self.Nsteps, len(self.axis_2D_1), len(self.axis_2D_2))

class DockArea(dockarea.DockArea, QObject):
    dock_signal = pyqtSignal()

    def __init__(self, temporary=False, home=None):
        super(DockArea, self).__init__(temporary, home)

    def moveDock(self, dock, position, neighbor):
        """
        Move an existing Dock to a new location.
        """
        ## Moving to the edge of a tabbed dock causes a drop outside the tab box
        if position in ['left', 'right', 'top', 'bottom'] and neighbor is not None and neighbor.container() is not None and neighbor.container().type() == 'tab':
            neighbor = neighbor.container()
        self.addDock(dock, position, neighbor)
        self.dock_signal.emit()

class ThreadCommand(object):
    """ | Micro class managing the thread commands.
        |
        | A thread command is composed of a string name defining the command to execute and an attribute list splitable making arguments of the called function.

        =============== =============
        **Attributes**  **Type**
        *command*       string
        *attributes*    generic list
        =============== =============

    """
    def __init__(self,command="",attributes=[]):
        self.command=command
        self.attributes=attributes

def elt_as_first_element(elt_list,match_word='Mock'):
    if elt_list!=[]:
        ind_elt=0
        for ind,elt in enumerate(elt_list):
            if 'Mock' in elt:
                ind_elt=ind
                break
        plugin_mock=elt_list[ind_elt]
        elt_list.remove(plugin_mock)
        plugins=[plugin_mock]
        plugins.extend(elt_list)
    else: plugins=[]
    return plugins

def send_string(socket, string):
    cmd_bytes, cmd_length_bytes = message_to_bytes(string)
    check_sended(socket, cmd_length_bytes)
    check_sended(socket, cmd_bytes)

def send_scalar(socket, data):
    """
    Convert it to numpy array then send the data type, the data_byte length and finally the data_bytes
    Parameters
    ----------
    data

    Returns
    -------

    """
    data = np.array([data])
    data_type = data.dtype.descr[0][1]
    data_bytes = data.tobytes()
    send_string(socket, data_type)
    check_sended(socket, len(data_bytes).to_bytes(4, 'big'))
    check_sended(socket, data_bytes)

def get_string(socket):
    string_len = get_int(socket)
    string = check_received_length(socket, string_len).decode()
    return string

def get_int(socket):
    data = int.from_bytes(check_received_length(socket, 4), 'big')
    return data

def get_scalar(socket):

    data_type = get_string(socket)
    data_len = get_int(socket)
    data_bytes = check_received_length(socket, data_len)

    data = np.frombuffer(data_bytes, dtype=data_type)[0]
    return data

def send_list(socket, data_list):

    check_sended(socket, len(data_list).to_bytes(4, 'big'))
    for data in data_list:

        if isinstance(data, np.ndarray):
            send_array(socket, data)

        elif isinstance(data, str):
            send_string(socket, data)

        elif isinstance(data, int) or isinstance(data, float):
            send_scalar(socket, data)

def get_list(socket, data_type):
    """
    Receive data from socket as a list
    Parameters
    ----------
    socket: the communication socket
    data_type: either 'string', 'scalar', 'array' depending which function has been used to send the list

    Returns
    -------

    """
    data = []
    list_len = get_int(socket)

    for ind in range(list_len):
        if data_type == 'scalar':
            data.append(get_scalar(socket))
        elif data_type == 'string':
            data.append(get_string(socket))
        elif data_type == 'array':
            data.append(get_array(socket))
    return data

def get_array(socket):
    data_type = get_string(socket)
    data_len = get_int(socket)
    Nrow = get_int(socket)
    Ncol = get_int(socket)
    data_bytes = check_received_length(socket, data_len)

    data = np.frombuffer(data_bytes, dtype=data_type)
    if Ncol != 0:
        data = data.reshape((Nrow, Ncol))
        # data=np.fliplr(data)  #because it gets inverted compared to digital...
    data = np.squeeze(data)  # in case one dimension is 1

    return data



def send_array(socket, data_array):
    """
    get data type as a string
    reshape array as 1D array and get number of rows and cols
    convert Data array as bytes
    send data type
    send data length
    send N rows
    send Ncols
    send data as bytes
    """
    data_type = data_array.dtype.descr[0][1]
    data_shape = data_array.shape
    Nrow = data_shape[0]
    if len(data_shape) > 1:
        Ncol = data_shape[1]
    else:
        Ncol = 1

    data = data_array.reshape(np.prod(data_shape))
    data_bytes = data.tobytes()

    send_string(socket, data_type)

    check_sended(socket, len(data_bytes).to_bytes(4, 'big'))
    check_sended(socket, Nrow.to_bytes(4, 'big'))
    check_sended(socket, Ncol.to_bytes(4, 'big'))
    check_sended(socket, data_bytes)

def message_to_bytes(message):
    """
    Convert a string to a byte array
    Parameters
    ----------
    message (str): message

    Returns
    -------
    Tuple consisting of the message converted as a byte array, and the length of the byte array, itself as a byte array of length 4
    """

    if not isinstance(message, bytes):
        message=message.encode()
    return message,len(message).to_bytes(4, 'big')

def check_sended(socket, data_bytes):
    """
    Make sure all bytes are sent through the socket
    Parameters
    ----------
    socket
    data_bytes

    Returns
    -------

    """
    sended = 0
    while sended < len(data_bytes):
        sended += socket.send(data_bytes[sended:])
    #print(data_bytes)


def check_received_length(sock,length):
    """
    Make sure all bytes (length) that should be received are received through the socket
    Parameters
    ----------
    sock
    length

    Returns
    -------

    """
    l=0
    data_bytes=b''
    while l<length:
        if l<length-4096:
            data_bytes_tmp=sock.recv(4096)
        else:
            data_bytes_tmp=sock.recv(length-l)
        l+=len(data_bytes_tmp)
        data_bytes+=data_bytes_tmp
    #print(data_bytes)
    return data_bytes


def find_in_path(path, mode):
    """
        Find the .py files in the given path directory

        =============== =========== ====================================
        **Parameters**    **Type**   **Description**
        *path*            string     The path to the directory to check
        =============== =========== ====================================

        Returns
        -------
        String list
            The list containing all the .py files in the directory.
    """
    plugins = []
    paths = os.listdir(path)
    for entry in paths:
        if mode in entry:
            if (mode == 'daq_move'):
                plugins.append(entry[9:-3])
            else:
                plugins.append(entry[13:-3])

    return plugins

def get_names_simple(mode):
    import pymodaq_plugins
    base_path = os.path.split(pymodaq_plugins.__file__)[0]

    if (mode == 'daq_move'):
        plugin_list = find_in_path(os.path.join(base_path, 'daq_move_plugins'), mode)
        plugins_import = elt_as_first_element(plugin_list, match_word='Mock')

    elif (mode == 'daq_0Dviewer'):
        plugin_list = find_in_path(os.path.join(base_path, 'daq_viewer_plugins', 'plugins_0D'), mode)
        plugins_import = elt_as_first_element(plugin_list, match_word='Mock')

    elif (mode == 'daq_1Dviewer'):
        plugin_list = find_in_path(os.path.join(base_path, 'daq_viewer_plugins', 'plugins_1D'), mode)
        plugins_import = elt_as_first_element(plugin_list, match_word='Mock')

    elif (mode == 'daq_2Dviewer'):
        plugin_list = find_in_path(os.path.join(base_path, 'daq_viewer_plugins', 'plugins_2D'), mode)
        plugins_import = elt_as_first_element(plugin_list, match_word='Mock')

    return plugins_import



def get_names(mode):
    """
        Get plugins names list from os dir command on plugins folder.
        The mode arguments specify the directory to list between DAQ_Move
        and DAQ_Viewer_XD

        =============== =========== ====================================================================
        **Parameters**    **Type**   **Description**
        *mode*            *string    The plugins directory to check between :
                                        * *DAQ_Move* : Check for DAQ_Move controllers plugins
                                        * *DAQ_Viewer_0D* : Chack for DAQ_Viewer\0D controllers plugins
                                        * *DAQ_Viewer_1D* : Chack for DAQ_Viewer\1D controllers plugins
                                        * *DAQ_Viewer_2D* : Chack for DAQ_Viewer\2D controllers plugins
        =============== =========== ====================================================================

        Returns
        -------
        String list
            The list containing all the present plugins names.

        See Also
        --------
        find_in_path
    """
    # liste=[]
    import pymodaq_plugins
    base_path = os.path.split(pymodaq_plugins.__file__)[0]
    # base_path=os.path.join(os.path.split(os.path.split(__file__)[0])[0],'plugins')
    if (mode == 'daq_move'):
        plugin_list = find_in_path(os.path.join(base_path, 'daq_move_plugins'), mode)
        plugins = elt_as_first_element(plugin_list, match_word='Mock')
        # check if modules are importable
        plugins_import = []
        for mod in plugins:
            try:
                importlib.import_module('.daq_move_' + mod, 'pymodaq_plugins.daq_move_plugins')
                plugins_import.append(mod)
            except:
                pass

        return plugins_import
    elif (mode == 'daq_0Dviewer'):
        plugin_list = find_in_path(os.path.join(base_path, 'daq_viewer_plugins', 'plugins_0D'), mode)
        plugins = elt_as_first_element(plugin_list, match_word='Mock')
        # check if modules are importable
        plugins_import = []
        for mod in plugins:
            try:
                importlib.import_module('.daq_0Dviewer_' + mod, 'pymodaq_plugins.daq_viewer_plugins.plugins_0D')
                plugins_import.append(mod)
            except:
                pass

        return plugins_import
    elif (mode == 'daq_1Dviewer'):
        plugin_list = find_in_path(os.path.join(base_path, 'daq_viewer_plugins', 'plugins_1D'), mode)
        plugins = elt_as_first_element(plugin_list, match_word='Mock')
        # check if modules are importable
        plugins_import = []
        for mod in plugins:
            try:
                importlib.import_module('.daq_1Dviewer_' + mod, 'pymodaq_plugins.daq_viewer_plugins.plugins_1D')
                plugins_import.append(mod)
            except:
                pass

        return plugins_import
    elif (mode == 'daq_2Dviewer'):
        plugin_list = find_in_path(os.path.join(base_path, 'daq_viewer_plugins', 'plugins_2D'), mode)
        plugins = elt_as_first_element(plugin_list, match_word='Mock')
        # check if modules are importable
        plugins_import = []
        for mod in plugins:
            try:
                importlib.import_module('.daq_2Dviewer_' + mod, 'pymodaq_plugins.daq_viewer_plugins.plugins_2D')
                plugins_import.append(mod)
            except:
                pass

        return plugins_import


# class EnumMeta (EnumMeta):

def make_enum(mode):
    """
        Custom class generator.
        Create a dynamic enum containing the plugins folder file names.

        Returns
        -------
        Instance of Enum
            The enum object representing loaded plugins
    """
    names=get_names(mode)
    values={}
    for i in range(0,len(names)):
        values.update({names[i]:i+1})
    meta=type(enum.Enum)
    bases=(enum.Enum,)
    dict=meta.__prepare__(names,bases)
    dict.update({'names': get_names})
    for key,value in values.items():
        dict[key]=value
    if(mode=='daq_move'):
        return meta(mode+'_Stage_type',bases,dict)
    else :
        return meta(mode+'_Type',bases,dict)





def check_modules(detectors, detectors_type, actuators):
    for ind_det, det in enumerate(detectors):
        if detectors_type[ind_det] == 'DAQ0D':
            if det not in get_names_simple('daq_0Dviewer'):
                raise Exception('Invalid actuator plugin')
        elif detectors_type[ind_det] == 'DAQ1D':
            if det not in get_names_simple('daq_1Dviewer'):
                raise Exception('Invalid actuator plugin')
        elif detectors_type[ind_det] == 'DAQ2D':
            if det not in get_names_simple('daq_2Dviewer'):
                raise Exception('Invalid actuator plugin')

    for act in actuators:
        if act not in get_names_simple('daq_move'):
            raise Exception('Invalid actuator plugin')

class PIDModelGeneric():
    params = []

    status_sig = pyqtSignal(ThreadCommand)
    actuators = []
    actuators_name = []
    detectors_type = [] # with entries either 'DAQ0D', 'DAQ1D' or 'DAQ2D'
    detectors = []
    detectors_name = []





    def __init__(self, pid_controller):
        self.pid_controller = pid_controller #instance of the pid_controller using this model
        self.settings = self.pid_controller.settings.child('models', 'model_params') #set of parameters
        self.data_names = None
        self.curr_output = None
        self.curr_input = None

    def update_detector_names(self):
        names = self.pid_controller.settings.child('main_settings', 'detector_modules').value()['selected']
        self.data_names = []
        for name in names:
            name = name.split('//')
            self.data_names.append(name)


    def update_settings(self, param):
        """
        Get a parameter instance whose value has been modified by a user on the UI
        To be overwritten in child class
        """
        if param.name() == '':
            pass

    def ini_model(self):
        pass

    def convert_input(self, measurements):
        """
        Convert the measurements in the units to be fed to the PID (same dimensionality as the setpoint)
        Parameters
        ----------
        measurements: (Ordereddict) Ordereded dict of object from which the model extract a value of the same units as the setpoint

        Returns
        -------
        float: the converted input

        """
        return 0

    def convert_output(self, output, dt):
        """
        Convert the output of the PID in units to be fed into the actuator
        Parameters
        ----------
        output: (float) output value from the PID from which the model extract a value of the same units as the actuator
        dt: (float) ellapsed time in seconds since last call
        Returns
        -------
        list: the converted output as a list (in case there are a few actuators)

        """
        #print('output converted')

        return [output]

def get_set_log_path():
    local_path = get_set_local_dir()
    log_path = os.path.join(local_path, 'logging')
    if not os.path.isdir(log_path):
        os.makedirs(log_path)
    return log_path

def get_set_pid_path():
    local_path = get_set_local_dir()
    pid_path = os.path.join(local_path, 'config_pid')
    if not os.path.isdir(pid_path):
        os.makedirs(pid_path)
    return pid_path


def zeros_aligned(n, align, dtype=np.uint32):
    """
    Get aligned memory array wih alignment align.
    Parameters
    ----------
    n: (int) length in dtype bytes of memory
    align: (int) memory alignment
    dtype: (numpy.dtype) type of the stored memory elements

    Returns
    -------

    """
    """
    
    """

    a = np.zeros(n + (align - 1), dtype=dtype)
    data_align = a.ctypes.data % align
    offset = 0 if data_align == 0 else (align - data_align)
    return a[offset: offset + n]

def cfunc(name, dll, result, *args):
    """build and apply a ctypes prototype complete with parameter flags

    Parameters
    ----------
    name: (str) function name in the dll
    dll: (ctypes.windll) dll object
    result : result is the type of the result (c_int,..., python function handle,...)
    args: list of tuples with 3 or 4 elements each like (argname, argtype, in/out, default) where argname is the
    name of the argument, argtype is the type, in/out is 1 for input and 2 for output, and default is an optional
    default value.

    Returns
    -------
    python function
    """
    atypes = []
    aflags = []
    for arg in args:
        atypes.append(arg[1])
        aflags.append((arg[2], arg[0]) + arg[3:])
    return CFUNCTYPE(result, *atypes)((name, dll), tuple(aflags))


def winfunc(name, dll, result, *args):
    """build and apply a ctypes prototype complete with parameter flags
    Parameters
    ----------
    name:(str) function name in the dll
    dll: (ctypes.windll) dll object
    result: result is the type of the result (c_int,..., python function handle,...)
    args: list of tuples with 3 or 4 elements each like (argname, argtype, in/out, default) where argname is the
    name of the argument, argtype is the type, in/out is 1 for input and 2 for output, and default is an optional
    default value.

    Returns
    -------
    python function
    """
    atypes = []
    aflags = []
    for arg in args:
        atypes.append(arg[1])
        aflags.append((arg[2], arg[0]) + arg[3:])
    return WINFUNCTYPE(result, *atypes)((name, dll), tuple(aflags))


def get_set_local_dir():
    if 'win32' in sys.platform:
        local_path = os.path.join(os.environ['HOMEDRIVE'] + os.environ['HOMEPATH'], 'pymodaq_local')
    else:
        local_path = os.path.join(os.environ['PATH'], 'pymodaq_local')

    if not os.path.isdir(local_path):
        os.makedirs(local_path)


    return local_path



def set_enable_recursive(children,enable=False):
    """
        | Set enable childs of chidren root argument with enable parameter value (False as default) calling recursively the method on children.
        |
        | Recursivity decreasing on children argument.

        =============== ===================== ==================================================
        **Parameters**    **Type**             **Description**
        *children*        settings tree node   The starting node of the (sub)tree to be treated
        *enable*          boolean              the default value to map
        =============== ===================== ==================================================

    """
    for child in children:
        if children==[]:
            return
        elif type(child) is QtWidgets.QSpinBox or type(child) is QtWidgets.QComboBox or type(child) is QtWidgets.QPushButton or type(child) is QtWidgets.QListWidget:
            child.setEnabled(enable)
        else:
            set_enable_recursive(child.children(),enable)


def find_file(string,extension):
    """
        Find .extension file list from the given list and the regular expression

        ============== ========== =======================================================
        **Parameters**   **Type**   **Description**
        *string*         string     raw splitted command result containing the file name
        *extension*      string     file extension (without .)
        ============== ========== =======================================================

        Returns
        -------
        string list
            The file list of the splitted os command result

        Examples
        --------
        >>> print(test_file)
        04/05/2018  11:55    <DIR>          .
        04/05/2018  11:55    <DIR>          ..
        04/05/2018  11:55    <DIR>          DAQ_Analysis
        04/05/2018  11:53             8ÿ758 find_and_replace.py
        03/05/2018  13:04             1ÿ327 find_py.py
        03/05/2018  13:25             3ÿ119 find_py_and_replace.py
        03/05/2018  15:47               619 find_words_in_line.py
        03/05/2018  16:02               524 replace_QtDesRess.py
        03/05/2018  13:20               142 test.py
        04/05/2018  11:53    <DIR>          __pycache__
                       6 fichier(s)           14ÿ489 octets
        >>> found_file=find_file(test_file,'py')
        >>> for i in range(0,len(found_file)):
        ...     print(found_file[i])
        ...
        find_and_replace.py
        find_py.py
        find_py_and_replace.py
        find_words_in_line.py
        replace_QtDesRess.py
        test.py
    """
    string_reg="([a-zA-Z0-9-_]*"+"\."+extension+")"
    regex=re.compile(string_reg,re.MULTILINE)
    ret=[]
    ret=re.findall(regex,string)
    re.purge()
    string_reg="[a-zA-Z0-9-_]*"
    regex=re.compile(string_reg,re.MULTILINE)
    for i in range(0,len(ret)):
        ret[i]=re.search(regex,ret[i]).group(0)
    return ret[:-1]


def recursive_find_files_extension(ini_path, ext, paths=[]):
    with os.scandir(ini_path) as it:
        for entry in it:
            if os.path.splitext(entry.name)[1][1:] == ext and entry.is_file():
                paths.append(entry.path)
            elif entry.is_dir():
                recursive_find_files_extension(entry.path, ext, paths)
    return paths






def nparray2Qpixmap(arr):
    result = QtGui.QImage(arr.data, arr.shape[1], arr.shape[0], QtGui.QImage.Format_RGB32)
    a=QtGui.QPixmap()
    a.convertFromImage(result)
    return a


def h5tree_to_QTree(h5file,base_node,base_tree_elt=None,pixmap_items=[]):
    """
        | Convert a loaded h5 file to a QTreeWidgetItem element structure containing two columns.
        | The first is the name of the h5 current node, the second is the path of the node in the h5 structure.
        |
        | Recursive function discreasing on base_node.

        ==================   ======================================== ===============================
        **Parameters**        **Type**                                 **Description**

          *h5file*            instance class File from tables module   loaded h5 file

          *base_node*         pytables h5 node                         parent node

          *base_tree_elt*     QTreeWidgetItem                          parent QTreeWidgetItem element
        ==================   ======================================== ===============================

        Returns
        -------
        QTreeWidgetItem
            h5 structure copy converted into QtreeWidgetItem structure.

        See Also
        --------
        h5tree_to_QTree

    """
    
    if base_tree_elt is None:
        base_tree_elt=QtWidgets.QTreeWidgetItem([base_node._v_name,"",base_node._v_pathname])
    for node in h5file.list_nodes(base_node):
        child=QtWidgets.QTreeWidgetItem([node._v_name,"",node._v_pathname])
        if 'pixmap' in node._v_attrs:
            pixmap_items.append(dict(node=node,item=child))
        if isinstance(node, tables.Group):
            h5tree_to_QTree(h5file,node,child,pixmap_items)

        base_tree_elt.addChild(child)

    return base_tree_elt,pixmap_items

def get_h5file_scans(h5file,path='/'):
    scan_list=[]
    for node in h5file.walk_nodes(path):
        if 'pixmap2D' in node._v_attrs:
            scan_list.append(dict(scan_name='{:s}_{:s}'.format(node._v_parent._v_name,node._v_name),path=node._v_pathname, data=node._v_attrs['pixmap2D']))

    return scan_list


def pixmap2ndarray(pixmap,scale=None):
    channels_count = 4
    image = pixmap.toImage()
    if scale==None:
        scale=[100,100]
    image=image.scaled(scale[0],scale[1],QtCore.Qt.KeepAspectRatio)

    #s = image.bits().asstring(image.width() * image.height() * channels_count)
    #arr = np.fromstring(s, dtype=np.uint8).reshape((image.height(), image.width(), channels_count)) 


    b = image.bits()
    # sip.voidptr must know size to support python buffer interface
    b.setsize(image.width() * image.height() * channels_count)
    arr = np.frombuffer(b, np.uint8).reshape((image.height(), image.width(), channels_count))
    return arr

def my_moment(x,y):
    dx=np.mean(np.diff(x))
    norm=np.sum(y)*dx
    m=[np.sum(x*y)*dx/norm]
    m.extend([np.sqrt(np.sum((x-m[0])**2*y)*dx/norm)])
    return m

def odd_even(x):
    """
		odd_even tells if a number is odd (return True) or even (return False)
		
		Parameters
		----------
		x: the integer number to test
		  
		Returns
		-------
		bool : boolean
    """
    if int(x)%2==0:
        bool=False
    else:
        bool=True
    return bool 

def set_scan_spiral(start_axis1,start_axis2,rmax,rstep):
    """Set a spiral scan of a 0D Data aquisition.

        =============== ========== ==========================================
        **Parameters**   **Type**   **Description**
        *start_axis1*    int        The starting value of the first sequence
        *start_axis2*    int        The starting value of the second sequence  
        *rmax*           int        The end point
        *rstep*          int        The value of one step
        =============== ========== ==========================================

        Returns
        -------
        (int,int list,int list,float list,float list,float list,list of 2 float lists) tuple
            The tuple containing :
             * the number of steps
             * the first axis indexes
             * the second axis indexes
             * the first axis with unique values
             * the second axis with unique values
             * the first axis
             * the second axis
             * the positions float values

        Examples
        --------

            >>> import daq_utils as Du
            >>> start_axis1,start_axis2=1,1
            >>> rmax=2
            >>> rstep=1
            >>> spiral_scan=Du.set_scan_spiral(start_axis1,start_axis2,rmax,rstep)
            >>> print(spiral_scan[0])       #The number of step
            25                                  
            >>> print(spiral_scan[1])       #The first distributed axis
            [-1  0  1  2  3]
            >>> print(spiral_scan[2])       #The second distributed axis
            [-1  0  1  2  3]
            >>> print(spiral_scan[3])       #The positions scalar list computed
            [[1, 1], [2, 1], [2, 2], [1, 2], [0, 2],
            [0, 1], [0, 0], [1, 0], [2, 0], [3, 0],
            [3, 1], [3, 2], [3, 3], [2, 3], [1, 3],
            [0, 3], [-1, 3], [-1, 2], [-1, 1], [-1, 0],
            [-1, -1], [0, -1], [1, -1], [2, -1], [3, -1]]
    """
    ind=0
    flag=True
    
    Nlin=np.trunc(rmax/rstep)    
    axis_1_indexes=[0]
    axis_2_indexes=[0]


    while flag:
        if odd_even(ind):
            step=1
        else:
            step=-1
        if flag:
            for ind_step in range(ind):
                axis_1_indexes.append(axis_1_indexes[-1]+step)
                axis_2_indexes.append(axis_2_indexes[-1])
                if len(axis_1_indexes)>=(2*Nlin+1)**2:
                    flag=False
                    break
        if flag:            
            for ind_step in range(ind):

                axis_1_indexes.append(axis_1_indexes[-1])
                axis_2_indexes.append(axis_2_indexes[-1]+step)
                if len(axis_1_indexes)>=(2*Nlin+1)**2:
                    flag=False
                    break
        ind+=1
    axis_1_indexes=np.array(axis_1_indexes,dtype=int)
    axis_2_indexes=np.array(axis_2_indexes,dtype=int)
    
    axis_1_unique=np.unique(axis_1_indexes)
    axis_1_unique=axis_1_unique.astype(float)
    axis_2_unique=np.unique(axis_2_indexes)
    axis_2_unique=axis_2_unique.astype(float)
    axis_1=np.zeros_like(axis_1_indexes,dtype=float)
    axis_2=np.zeros_like(axis_2_indexes,dtype=float)
    
    positions=[]
    for ind in range(len(axis_1)):
        axis_1[ind]=axis_1_indexes[ind]*rstep+start_axis1
        axis_2[ind]=axis_2_indexes[ind]*rstep+start_axis2
        positions.append([axis_1[ind],axis_2[ind]])

    for ind in range(len(axis_1_unique)):
        axis_1_unique[ind]=axis_1_unique[ind]*rstep+start_axis1
        axis_2_unique[ind]=axis_2_unique[ind]*rstep+start_axis2

 
    axis_1_indexes=axis_1_indexes-np.min(axis_1_indexes)
    axis_2_indexes=axis_2_indexes-np.min(axis_2_indexes)


    Nsteps=len(positions)
    return ScanParameters(Nsteps,axis_1_indexes,axis_2_indexes,axis_1_unique,axis_2_unique,positions)

def linspace_step(start,stop,step):
    """
        Compute a regular linspace_step distribution from start to stop values.

        =============== =========== ======================================
        **Parameters**    **Type**    **Description**
        *start*            scalar      the starting value of distribution
        *stop*             scalar      the stopping value of distribution
        *step*             scalar      the length of a distribution step
        =============== =========== ======================================

        Returns
        -------

        scalar array
            The computed distribution axis as an array.

        Examples
        --------
        >>> import DAQ_utils as Du
        >>> import numpy as np
        >>>    #arguments initializing
        >>> start=0
        >>> stop=5
        >>> step=0.25
        >>> linspace_distribution=Du.linspace_step(start,stop,step)
        >>> print(linspace_distribution)
        >>>    #computed distribution
        [ 0.    0.25  0.5   0.75  1.    1.25  1.5   1.75  2.    2.25  2.5   2.75
          3.    3.25  3.5   3.75  4.    4.25  4.5   4.75  5.  ]        
    """
    tmp=start
    out=np.array([tmp])    
    if step>=0:
        while (tmp<=stop):
            tmp=tmp+step
            out=np.append(out,tmp)
    else:
        while (tmp>=stop):
            tmp=tmp+step
            out=np.append(out,tmp)
    return out[0:-1]



def set_scan_linear(start_axis1,start_axis2,stop_axis1,stop_axis2,step_axis1,step_axis2,back_and_force=False):
    """
        Set a linear scan of a 0D Data aquisition.
        The positions scalar list is computed by a Cartesian product of the first distributed axis and the second one.
        
        The result size is composed by :
        * a single integer representing the number of step
        * a n items integer array representing the first distributed axis
        * a k items integer array representing the second distributed axis
        * a n*k items containing the combinaisons of the first and the second axis distribution.

        ================ ========== =============================================
        **Parameters**    **Type**   **Description**
        *start_axis1*     scalar     The starting value of the first sequence
        *start_axis2*     scalar     The starting value of the second sequence
        *stop_axis1*      scalar     The end point of the first sequence
        *stop_axis2*      scalar     The end point of the second sequence
        *step_axis1*      float      The value of one step of the first sequence
        *step_axis2*     float      The value of one step of the second sequence
        *back_and_force*  boolean    ???
        ================ ========== =============================================


        Returns
        -------
        (int,float list,float list,scalar list) tuple
            The tuple containing:
             * The number of step
             * The first distributed axis
             * The second distributed axis
             * The positions scalar list computed

        Examples
        --------
        ..doctest::

            >>> import DAQ_utils as Du
            >>> start_axis1,start_axis2=1,1
            >>> stop_axis1,stop_axis2=3,3
            >>> step_axis1,step_axis2=1,1
            >>> linear_scan=Du.set_scan_linear(start_axis1,start_axis2,stop_axis1,stop_axis2,step_axis1,step_axis2)
            >>> print(linear_scan[0])       #The number of step
            9
            >>> print(linear_scan[1])       #The first distributed axis
            [1 2 3]
            >>> print(linear_scan[2])       #The second distributed axis
            [1 2 3]
            >>> print(linear_scan[3])       #The positions scalar list computed
            [[1, 1], [1, 2], [1, 3], [2, 1], [2, 2], [2, 3], [3, 1], [3, 2], [3, 3]]
    """
    axis_1_unique=linspace_step(start_axis1,stop_axis1,step_axis1)
    axis_2_unique=linspace_step(start_axis2,stop_axis2,step_axis2)
    positions=[]
    axis_1_indexes=[]
    axis_2_indexes=[]
    axis_1=[]
    axis_2=[]   
    for ind_x,pos1 in enumerate(axis_1_unique):
        if back_and_force:
            for ind_y,pos2 in enumerate(axis_2_unique):
                if not odd_even(ind_x):
                    positions.append([pos1,pos2])
                    axis_1.append(pos1)
                    axis_2.append(pos2)
                    axis_1_indexes.append(ind_x)
                    axis_2_indexes.append(ind_y)
                else:
                    positions.append([pos1,axis_2_unique[len(axis_2_unique)-ind_y-1]])
                    axis_1.append(pos1)
                    axis_2.append(axis_2_unique[len(axis_2_unique)-ind_y-1])
                    axis_1_indexes.append(ind_x)
                    axis_2_indexes.append(len(axis_2_unique)-ind_y-1)   
        else:
            for ind_y,pos2 in enumerate(axis_2_unique):
                axis_1.append(pos1)
                axis_2.append(pos2)
                positions.append([pos1,pos2])
                axis_1_indexes.append(ind_x)
                axis_2_indexes.append(ind_y)


    Nsteps=len(positions)
    return ScanParameters(Nsteps,axis_1_indexes,axis_2_indexes,axis_1_unique,axis_2_unique,positions)


def set_scan_random(start_axis1,start_axis2,stop_axis1,stop_axis2,step_axis1,step_axis2):
    scan_parameters = set_scan_linear(start_axis1, start_axis2, stop_axis1, stop_axis2, step_axis1, step_axis2, back_and_force=False)

    positions_shuffled=scan_parameters.positions[:]
    np.random.shuffle(positions_shuffled)
    axis_1_indexes=[]
    axis_2_indexes=[]

    for pos in positions_shuffled:
        axis_1_indexes.append(np.where(scan_parameters.axis_2D_1==pos[0])[0][0])
        axis_2_indexes.append(np.where(scan_parameters.axis_2D_2==pos[1])[0][0])


    Nsteps = len(scan_parameters.positions)
    return ScanParameters(Nsteps,axis_1_indexes,axis_2_indexes,scan_parameters.axis_2D_1,scan_parameters.axis_2D_2,
                          positions_shuffled)

def set_param_from_param(param_old,param_new):
    """
        Walk through parameters children and set values using new parameter values.
    """
    for child_old in param_old.children():
        try:
            path=param_old.childPath(child_old)
            child_new=param_new.child(*path)
            param_type=child_old.type()

            if 'group' not in param_type: #covers 'group', custom 'groupmove'...
                try:
                    if 'list' in param_type:#check if the value is in the limits of the old params (limits are usually set at initialization)
                        if child_new.value() not in child_old.opts['limits']:
                            child_old.opts['limits'].append(child_new.value())

                        child_old.setValue(child_new.value())
                    elif 'str' in param_type or 'browsepath' in param_type or 'text' in param_type:
                        if child_new.value()!="":#to make sure one doesnt overwrite something
                            child_old.setValue(child_new.value())
                    else:
                        child_old.setValue(child_new.value())
                except Exception as e:
                    print(str(e))
            else:
                set_param_from_param(child_old,child_new)
        except Exception as e:
            print(str(e))

def find_part_in_path_and_subpath(base_dir,part='',create=False):
    """
        Find path from part time.

        =============== ============ =============================================
        **Parameters**  **Type**      **Description**
        *base_dir*      Path object   The directory to browse
        *part*          string        The date of the directory to find/create
        *create*        boolean       Indicate the creation flag of the directory
        =============== ============ =============================================

        Returns
        -------
        Path object
            found path from part

        Examples
        --------
        ..doctest::

            >>> import DAQ_utils as Du
            >>> import pathlib as pl
            >>> base_dir=pl.Path("") #Getting the current path
            >>> print(base_dir)
            .
            >>> path=Du.find_part_in_path_and_subpath(base_dir,"2018",True)
            >>> print(path)       #Path of created directory "2018"
            2018
            >>> path=Du.find_part_in_path_and_subpath(base_dir,"2017",False)
            >>> print(path)       #Path is none since "2017" dir doesn't exist
            None
            >>> path=Du.find_part_in_path_and_subpath(base_dir,"2018",False)
            >>> print(path)       #Path of directory "2018"
            2018
    """
    found_path=None
    if part in base_dir.parts: #check if current year is in the given base path
        if base_dir.name==part:
            found_path=base_dir
        else:
            for ind in range(len(base_dir.parts)):
                tmp_path=base_dir.parents[ind]
                if tmp_path.name==part:
                    found_path=base_dir.parents[ind]
                    break
    else:#if not check if year is in the subfolders
        subfolders_year_name=[x.name for x in base_dir.iterdir() if x.is_dir()]
        subfolders_found_path=[x for x in base_dir.iterdir() if x.is_dir()]
        if part not in subfolders_year_name:
            if create:
                found_path=base_dir.joinpath(part)
                found_path.mkdir()
            else:
                found_path = base_dir
        else:
            ind_path=subfolders_year_name.index(part)
            found_path=subfolders_found_path[ind_path]
    return found_path

def set_current_scan_path(base_dir,base_name='Scan',update_h5=False,next_scan_index=0,create_scan_folder = False, create_dataset_folder=True):
    """
        Set the path of the current scan and create associated directory tree.
        As default :

        Year/Date/Dataset_Date_ScanID/ScanID


        =============== ============ =====================================
        **Parameters**  **Type**      **Description**
        base_dir        Path object   The base directory
        base_name       string        Name of the current scan
        update_h5       boolean       1/0 to update the associated h5 file
        =============== ============ =====================================

        Returns
        -------
        scan_path 
            indexed base_name, new folder indexed from base name at the day

        See Also
        --------
        DAQ_utils.find_part_in_path_and_subpath

        Examples
        --------
        ..doctest::

            >>> import DAQ_utils as Du
            >>> import pathlib as pl
            >>> base_dir=pl.Path("") #Getting the current path
            >>>
            >>> current_scan_path=Du.set_current_scan_path(base_dir)
            >>>  #Function call with default name
            >>> print(current_scan_path[0])       #The full scan path
            2018\20180424\Dataset_20180424_000\Scan000
            >>> print(current_scan_path[1])       #The indexed base name
            Scan000
            >>> print(current_scan_path[2])       #The dataset_path
            2018\20180424\Dataset_20180424_000
            >>>
            >>> current_scan_path=Du.set_current_scan_path(base_dir,'Specific name')
            >>> #Function call with a specific name
            >>> print(current_scan_path[0])       #The full scan path
            2018\20180424\Dataset_20180424_000\Specific name000
            >>> print(current_scan_path[1])       #The indexed base name
            Specific name000
            >>> print(current_scan_path[2])       #The dataset_path
            2018\20180424\Dataset_20180424_000
    """
    base_dir=Path(base_dir)
    date=datetime.date.today()

    year_path=find_part_in_path_and_subpath(base_dir,part=str(date.year),create=True)# create directory of the year if it doen't exist and return it
    day_path=find_part_in_path_and_subpath(year_path,part=date.strftime('%Y%m%d'),create=True)# create directory of the day if it doen't exist and return it

    date=datetime.date.today()
    dataset_base_name=date.strftime('Dataset_%Y%m%d')
    dataset_paths=sorted([path for path in day_path.glob(dataset_base_name+"*") if path.is_dir()])
    if dataset_paths==[]:
        ind_dataset=0
    else:
        if update_h5:
            ind_dataset=int(dataset_paths[-1].name.partition(dataset_base_name+"_")[2])+1
        else:
            ind_dataset=int(dataset_paths[-1].name.partition(dataset_base_name+"_")[2])
    dataset_path=find_part_in_path_and_subpath(day_path,part=dataset_base_name+"_{:03d}".format(ind_dataset),create=create_dataset_folder)

    scan_paths=sorted([path for path in dataset_path.glob(base_name+'*') if path.is_dir()])
    # if scan_paths==[]:
    #     ind_scan=0
    # else:
    #     if list(scan_paths[-1].iterdir())==[]:
    #         ind_scan=int(scan_paths[-1].name.partition(base_name)[2])
    #     else:
    #         ind_scan=int(scan_paths[-1].name.partition(base_name)[2])+1
    ind_scan = next_scan_index

    scan_path=find_part_in_path_and_subpath(dataset_path,part=base_name+'{:03d}'.format(ind_scan),create=create_scan_folder)
    return scan_path,base_name+'{:03d}'.format(ind_scan),dataset_path


def select_file(start_path=None,save=True, ext=None):
    """
        Save or open a file with Qt5 file dialog, to be used within an Qt5 loop.

        =============== ======================================= ===========================================================================
        **Parameters**     **Type**                              **Description**

        *start_path*       Path object or str or None, optional  the path Qt5 will open in te dialog
        *save*             bool, optional                        * if True, a savefile dialog will open in order to set a savefilename
                                                                 * if False, a openfile dialog will open in order to open an existing file
        *ext*              str, optional                         the extension of the file to be saved or opened
        =============== ======================================= ===========================================================================

        Returns
        -------
        Path object
            the Path object pointing to the file

        Examples
        --------
        >>>from PyQt5 import QtWidgets
        >>>from PyMoDAQ.DAQ_Utils.DAQ_utils import select_file
        >>>import sys
        >>>app = QtWidgets.QApplication(sys.argv)
        >>>    #Open a save windows
        >>>select_file(start_path="C:/test",save=False,ext='h5')
        >>>sys.exit(app.exec_())

    """
    if ext is None:
        ext='*'
    if not save:
        if type(ext)!=list:
            ext=[ext]
        
        filter="Data files ("
        for ext_tmp in ext:
            filter+='*.'+ext_tmp+" "
        filter+=")"
    if start_path is not None:
        if not isinstance(start_path, str):
            start_path=str(start_path)
    if save:
        fname = QtWidgets.QFileDialog.getSaveFileName(None, 'Enter a .'+ext+' file name',start_path,ext+" file (*."+ext+")")
    else:
        fname=QtWidgets.QFileDialog.getOpenFileName(None, 'Select a file name',start_path,filter)

    fname=fname[0]
    if fname != '': #execute if the user didn't cancel the file selection
        fname=Path(fname)
        if save:
            parent=fname.parent
            filename=fname.stem
            fname=parent.joinpath(filename+"."+ext) #forcing the right extension on the filename
    return fname #fname is a Path object

def find_index(x,threshold):
    """
    find_index finds the index ix such that x(ix) is the closest from threshold
    
    Parameters
    ----------
    x : vector
    threshold : list of scalar

    Returns
    -------
    out : list of 2-tuple containing ix,x[ix]
            out=[(ix0,xval0),(ix1,xval1),...]
    
    Examples
    --------
    >>> import array_manipulation as am
    >>> import numpy as np
    >>>    #vector creation
    >>> x_vector=np.array([1,2,3])
    >>> index=am.find_index(x_vector,4)
    >>>    #the nearest index from threshold
    >>> print(index)
    >>>    #ix=2, x[2]=3  , threshold=4
    [(2, 3)]
    >>>
    >>> x_vector=np.array([1,2,3,4,5,6,7,8,9,10])
    >>> index=am.find_index(x_vector,[3.8,7.5,12.])
    >>>    #the nearest indexs from threshold list [3.8,7.5,12.]
    >>> print(index)
    >>>    #ix0=3, x[3]=4  , threshold=3.8
    >>>    #ix1=6, x[6]=7  , threshold=7.5
    >>>    #ix2=9, x[9]=10 , threshold=12.0
    [(3, 4), (6, 7), (9, 10)]
    """
        
    if np.isscalar(threshold):
        threshold=[threshold]
    out=[]
    for value in threshold:
        ix=int(np.argmin(np.abs(x-value))) 
        out.append((ix,x[ix]))
    return out

def gauss1D(x,x0,dx,n=1):
    """
    compute the gaussian function along a vector x, centered in x0 and with a
    FWHM i intensity of dx. n=1 is for the standart gaussian while n>1 defines
    a hypergaussian

    =============== =========== ============================================================
    **Parameters**    **Type**    **Description**
    *x*               vector      vector
    *x0*              float       the central position of the gaussian
    *dx*              float       the FWHM of the gaussian
    *n*               float       define hypergaussian, n=1 by default for regular gaussian
    =============== =========== ============================================================     

    Returns
    -------
    out : vector
      the value taken by the gaussian along x axis

    Examples
    --------
    >>> import DAQ_utils as Du
    >>> import numpy as np                
    >>> x=np.array([0,1,2,3,4])            #argument initializing
    >>> x0=2
    >>> dx=0.5
    >>> gaussian_distribution=Du.gauss1D(x,x0,dx)
    >>> print(gaussian_distribution)       #The computed gaussian distribution
    [  2.32830644e-10   3.90625000e-03   1.00000000e+00   3.90625000e-03
       2.32830644e-10]    
    """
    out=np.exp(-2*np.log(2)**(1/n)*(((x-x0)/dx))**(2*n))
    return out

def gauss2D(x,x0,dx,y,y0,dy,n=1,angle=0):
    """
    compute the 2D gaussian function along a vector x, centered in x0 and with a
    FWHM in intensity of dx and smae along y axis. n=1 is for the standart gaussian while n>1 defines
    a hypergaussian. optionally rotate it by an angle in degree

    Parameters
    ----------
    x : vector
    x0 : a float: the central position of the gaussian
    dx : a float :the FWHM of the gaussian
    n=1 : a float to define hypergaussian, n=1 by default for regular gaussian        
    angle=0 : a float to rotate main axes
    
    Returns
    -------
    out : vector
      the value taken by the gaussian along x axis
    """
    if angle==0:
        data=np.transpose(np.outer(gauss1D(x,x0,dx,n),gauss1D(y,y0,dy,n)))

    else:

        theta = np.radians(angle)
        c, s = np.cos(theta), np.sin(theta)
        R = np.array(((c,-s), (s, c)))
        (x0r,y0r)=tuple(R.dot(np.array([x0,y0])))
        
        data=np.zeros((len(y),len(x)))
        
        for indx,xtmp in enumerate(x):
            for indy,ytmp in enumerate(y):
                rotatedvect=R.dot(np.array([xtmp,ytmp]))
                data[indy,indx]=np.exp(-2*np.log(2)**(1/n)*((rotatedvect[0]-x0r)/dx)**(2*n))*np.exp(-2*np.log(2)**(1/n)*((rotatedvect[1]-y0r)/dy)**(2*n))
                  
    return data
    
def ftAxis_time(Npts,time_max):
    """
    Given two numbers Npts,omega_max, return two vectors spanning the temporal
    and spectral range. They are related by Fourier Transform

    =============== =========== ===========================================================================
    **Parameters**    **Type**    **Description**
    *Npts*            int          A number of points defining the length of both grids
    *time_max*       float         The maximum circular frequency in the spectral domain. its unit defines
                                   the temporal units. ex: omega_max in rad/fs implies time_grid in fs
    =============== =========== ===========================================================================

    Returns
    -------
    omega_grid : vector
      The spectral axis of the FFT

    time_grid : vector
      The temporal axis of the FFT

    Example
    -------
    >>> (omega_grid, time_grid)=ftAxis(Npts,omega_max)
    ...
    """
    dT=time_max/Npts
    omega_max=(Npts-1)/2*2*np.pi/time_max
    omega_grid = np.linspace(-omega_max,omega_max,Npts)
    time_grid = dT*np.linspace(-(Npts-1)/2,(Npts-1)/2,Npts)
    return omega_grid, time_grid
    
    
def ft(x,dim=0):
    out=np.fft.fftshift(np.fft.fft(np.fft.fftshift(x,axes=dim),axis=dim),axes=dim)
    return out
    
def ift(x,dim=0):
    out=np.fft.fftshift(np.fft.ifft(np.fft.fftshift(x,axes=dim),axis=dim),axes=dim)
    return out


def ftAxis(Npts, omega_max):
    """
    Given two numbers Npts,omega_max, return two vectors spanning the temporal
    and spectral range. They are related by Fourier Transform

    Parameters
    ----------
    Npts : number
      A number of points defining the length of both grids
    omega_max : number
      The maximum circular frequency in the spectral domain. its unit defines
      the temporal units. ex: omega_max in rad/fs implies time_grid in fs

    Returns
    -------
    omega_grid : vector
      The spectral axis of the FFT

    time_grid : vector
      The temporal axis of the FFT

    Example
    -------
    >>> (omega_grid, time_grid)=ftAxis(Npts,omega_max)
    ...
    """
    dT = 2 * np.pi / (2 * omega_max)
    omega_grid = np.linspace(-omega_max, omega_max, Npts)
    time_grid = dT * np.linspace(-(Npts - 1) / 2, (Npts - 1) / 2, Npts)
    return omega_grid, time_grid


def ftAxis_time(Npts, time_max):
    """
    Given two numbers Npts,omega_max, return two vectors spanning the temporal
    and spectral range. They are related by Fourier Transform

    Parameters
    ----------
    Npts : number
      A number of points defining the length of both grids
    omega_max : number
      The maximum circular frequency in the spectral domain. its unit defines
      the temporal units. ex: omega_max in rad/fs implies time_grid in fs

    Returns
    -------
    omega_grid : vector
      The spectral axis of the FFT

    time_grid : vector
      The temporal axis of the FFT

    Example
    -------
    >>> (omega_grid, time_grid)=ftAxis(Npts,omega_max)
    ...
    """
    dT = time_max / Npts
    omega_max = (Npts - 1) / 2 * 2 * np.pi / time_max
    omega_grid = np.linspace(-omega_max, omega_max, Npts)
    time_grid = dT * np.linspace(-(Npts - 1) / 2, (Npts - 1) / 2, Npts)
    return omega_grid, time_grid


def ft(x, dim=0):
    out = np.fft.fftshift(np.fft.fft(np.fft.fftshift(x, axes=dim), axis=dim), axes=dim)
    return out


def ft2(x, dim=None):
    out = np.fft.fftshift(np.fft.fft2(np.fft.fftshift(x, axes=dim)), axes=dim)
    return out


def ift2(x, dim=0):
    out = np.fft.fftshift(np.fft.ifft2(np.fft.fftshift(x, axes=dim)), axes=dim)
    return out







if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    form = QtWidgets.QWidget();
    file_path=".//test.h5"
    h5file=tables.open_file(file_path)
    prog = H5Browser(form,h5file)
    form.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
"""
Created on Mon Dec  4 10:59:53 2017

@author: Weber

"""
import sys
import PyQt5
from PyQt5 import QtWidgets,QtGui
from PyQt5.QtCore import pyqtSlot, pyqtSignal, QLocale, Qt, QDate, QDateTime, QTime, QByteArray
from pyqtgraph.widgets import ColorButton, SpinBox
import pyqtgraph.parametertree.parameterTypes as pTypes
from pyqtgraph.parametertree import Parameter, ParameterItem
from pyqtgraph.parametertree.Parameter import registerParameterType
#from PyMoDAQ.daq_utils.plotting.select_item_tolist_main import Select_item_tolist_simpler
from pymodaq.daq_utils.daq_utils import scroll_log, scroll_linear
from collections import OrderedDict
from decimal import Decimal as D

from pymodaq.daq_utils.plotting.qled import QLED
import xml.etree.ElementTree as ET

from pathlib import Path
import numpy as np
import os


def get_param_path(param):
    path = []
    par_tmp = param
    while par_tmp.parent() is not None:
        path.append(par_tmp.name())
        par_tmp = param.parent()
    return par_tmp[-1::-1]

#%% with attribute
def walk_parameters_to_xml(parent_elt=None,param=None):
    """
        To convert a parameter object (and children) to xml data tree.

        =============== ================================ ==================================
        **Parameters**   **Type**                         **Description**

        *parent_elt*     XML element                      the root element
        *param*          instance of pyqtgraph parameter  Parameter object to be converted
        =============== ================================ ==================================

        Returns
        -------
        XML element : parent_elt
            XML element with subelements from Parameter object

        Examples
        -------
        >>> from pyqtgraph.parametertree import Parameter, ParameterItem
        >>> import xml.etree.ElementTree as ET
        >>> params = [{title': 'Scan2D settings', 'name': 'scan2D_settings', 'type': 'group', 'children': [
                        {'title': 'Scan type:','name': 'scan2D_type', 'type': 'list', 'values': ['Spiral','Linear', 'back&forth'],'value': 'Spiral'},
                        {'title': 'Rstep:','name': 'Rstep_2d', 'type': 'float', 'value': 1., 'visible':True},
                        {'title': 'Rmax:','name': 'Rmax_2d', 'type': 'float', 'value': 10., 'visible':True}
                        ]}]
        >>> settings=Parameter.create(name='Settings', type='group', children=params)
        >>> param_type=settings.type()
        >>> base_elt=ET.Element(settings.name(),title=str(settings.opts['title']),type=param_type)

        >>> XML_elt=walk_parameters_to_xml(param=settings)
        >>> tree=ET.ElementTree(XML_elt)
        >>> tree.write('settings.xml')


        See Also
        --------
        walk_parameters_to_xml

    """

    if type(param) is None:
        raise TypeError('No valid param input')

    if parent_elt is None:
        opts = dict_from_param(param)
        parent_elt = ET.Element(param.name(), **opts)
        param_type = str(param.type())
        if 'group' not in param_type:  # covers 'group', custom 'groupmove'...
            add_text_to_elt(parent_elt, param)

    params_list=param.children()
    for param in params_list:
        opts = dict_from_param(param)
        elt = ET.Element(param.name(), **opts)
        param_type = str(param.type())
        if 'group' not in param_type: #covers 'group', custom 'groupmove'...
            add_text_to_elt(elt, param)
        else:
            walk_parameters_to_xml(elt, param)

        parent_elt.append(elt)
    return parent_elt


def add_text_to_elt(elt, param):
    param_type = str(param.type())
    if param_type == 'bool' or param_type == 'bool_push':
        if param.value():
            text = '1'
        else:
            text = '0'
    elif param_type == 'itemselect':
        if param.value() is not None:
            elt.set('all_items',
                    str(param.value()['all_items']))  # use list(eval(val_str[1:-1])) to get back a list of strings
            text = str(param.value()['selected'])  # use list(eval(val_str[1:-1])) to get back a list of strings
        else:
            text = str(None)
    elif param_type == 'color':
        text = str([param.value().red(), param.value().green(), param.value().blue(), param.value().alpha()])
    elif param_type == 'list':
        if isinstance(param.value(), str):
            text = "str('{}')".format(param.value())
        elif isinstance(param.value(), int):
            text = 'int({})'.format(param.value())
        elif isinstance(param.value(), float):
            text = 'float({})'.format(param.value())
        else:
            str(param.value())
    else:
        text = str(param.value())
    elt.text = text

def dict_from_param(param):
    opts = dict([])
    param_type = str(param.type())
    opts.update(dict(type=param_type))
    title = param.opts['title']
    if title is None:
        title = param.name()
    opts.update(dict(title=title))

    if param.opts['visible']:
        visible = '1'
    else:
        visible = '0'
    opts.update(dict(visible=visible))

    if param.opts['removable']:
        removable = '1'
    else:
        removable = '0'
    opts.update(dict(removable=removable))

    if param.opts['readonly']:
        readonly = '1'
    else:
        readonly = '0'
    opts.update(dict(readonly=readonly))

    if 'detlist' in param.opts:
        detlist = str(param.opts['detlist'])
        opts.update(dict(detlist=detlist))

    if 'movelist' in param.opts:
        movelist = str(param.opts['movelist'])
        opts.update(dict(movelist=movelist))

    if 'show_pb' in param.opts:
        if param.opts['show_pb']:
            show_pb = '1'
        else:
            show_pb = '0'
        opts.update(dict(show_pb=show_pb))

    return opts

def parameter_to_xml_string(param):
    """
        Convert  the given parameter to XML string.

        =============== ================================ ===============================
        **Parameters**    **Type**                        **Description**
        paramm           instance of pyqtgraph parameter  The parameter to be converted
        =============== ================================ ===============================

        Returns
        -------
        string
            The converted string XML element.

        See Also
        --------
        walk_parameters_to_xml

        Examples
        --------

        >>> import custom_parameter_tree as cpt
        >>> from pyqtgraph.parametertree import Parameter
        >>>    #Create an instance of Parameter
        >>> settings=Parameter(name='settings')  
        >>> converted_xml=cpt.parameter_to_xml_string(settings)
        >>>    # The converted Parameter
        >>> print(converted_xml)                 
        b'<settings title="settings" type="None" />'
    """
    xml_elt=walk_parameters_to_xml(param=param)
    return ET.tostring(xml_elt)

def parameter_to_xml_file(param,filename):
    """
        Convert the given parameter to XML element and update the given XML file.

        =============== ================================= ================================
        **Parameters**    **Type**                          **Description**
        *param*           instance of pyqtgraph parameter   the parameter to be added
        *filename*        string                            the filename of the XML file
        =============== ================================= ================================

        See Also
        --------
        walk_parameters_to_xml  

        Examples
        --------
        >>> import custom_parameter_tree as cpt
        >>> from pyqtgraph.parametertree import Parameter
        >>> import pathlib as pl
        >>>    #Creating function parameters
        >>> settings=Parameter(name='settings')
        >>> filename="my_xml_file"
        >>> cpt.parameter_to_xml_file(settings,filename)
        >>>    #Verifiy the integrity of the converted xml file
        >>> check=open("my_xml_file.xml","r")
        >>> print(check.read())
        <settings title="settings" type="None" />      
    """
    fname=Path(filename)
    parent=fname.parent
    filename=fname.stem
    fname=parent.joinpath(filename+".xml") #forcing the right extension on the filename
    xml_elt=walk_parameters_to_xml(param=param)
    tree=ET.ElementTree(xml_elt)
    tree.write(str(fname))



def walk_xml_to_parameter(params=[],XML_elt=None):
    """ To convert an XML element (and children) to dict enabling creation of parameter object.

        =============== ================== =======================================
        **Parameters**   **Type**            **Description**

        *params*         dictionnary list    the list to create parameter object
        *XML_elt*        XML object          the XML object to be converted
        =============== ================== =======================================

        Returns
        -------
        params : dictionnary list
            list of dict to create parameter object

        Examples
        -------
        >>> from pyqtgraph.parametertree import Parameter, ParameterItem
        >>> import xml.etree.ElementTree as ET
        >>> tree = ET.parse('text_bis.xml')
        >>> root = tree.getroot()
        >>> params=walk_xml_to_parameter(XML_elt=root)
        >>> settings_xml=Parameter.create(name='Settings XML', type='group', children=params)
        >>> settings=Parameter.create(name='Settings', type='group', children=params)

        See Also
        --------
        walk_parameters_to_xml          
    """
    try:
        if type(XML_elt) is not ET.Element:
            raise TypeError('not valid XML element')


        elts = XML_elt.getchildren()
        if len(elts) == 0:
            param_dict = elt_to_dict(XML_elt)
            param_type = XML_elt.get('type')

            if 'group' not in param_type:  # covers 'group', custom 'groupmove'...
                set_txt_from_elt(XML_elt, param_dict)
            params.append(param_dict)

        for el in elts:
            param_dict = elt_to_dict(el)
            param_type = el.get('type')

            if 'group' not in param_type: #covers 'group', custom 'groupmove'...
                set_txt_from_elt(el, param_dict)
            else:
                subparams=[]
                param_dict.update(dict(children=walk_xml_to_parameter(subparams,el)))

                param_dict.update(dict(name=el.tag))

            params.append(param_dict)
    except Exception as e: #to be able to debug when there's an issue
        raise e
    return params


def set_txt_from_elt(el, param_dict):
    """
    get the value of the parameter from the text value of the xml element
    Parameters
    ----------
    el: xml element
    param_dict: dictionnary from which the parameter will be constructed

    """
    val_text = el.text
    param_type = el.get('type')

    if param_type == 'float':
        param_value = float(val_text)
    elif param_type == 'int':
        param_value = int(val_text)
    elif param_type == 'slide':
        param_value = float(val_text)
    elif param_type == 'itemselect':
        if val_text == 'None':
            param_value = dict(all_items=[], selected=[])
        else:
            param_value = dict(all_items=eval(el.get('all_items', val_text)), selected=eval(val_text))
    elif param_type == 'bool':
        param_value = bool(int(val_text))
    elif param_type == 'bool_push':
        param_value = bool(int(val_text))
    elif param_type == 'led':
        param_value = bool(val_text)
    elif param_type == 'date_time':
        param_value = eval(val_text)
    elif param_type == 'date':
        param_value = eval(val_text)
    elif param_type == 'table':
        param_value = eval(val_text)
    elif param_type == 'color':
        param_value = QtGui.QColor(*eval(val_text))
    elif param_type == 'list':
        try:
            param_value = eval(val_text)
        except:
            param_value = val_text  # for back compatibility
    else:
        param_value = val_text
    param_dict.update(dict(value=param_value))

    if param_type == 'list':
        param_dict.update(dict(values=[param_value]))


def elt_to_dict(el):
    param = dict([])

    # name=el.tag, title=title, type=param_type, value=param_value, values=[param_value],
    #              visible=visible, removable=removable, readonly=readonly, show_pb=show_pb)
    param.update(dict(name=el.tag))
    param_type=el.get('type')
    param.update(dict(type=param_type))

    title=el.get('title')
    if title=='None':
        title=el.tag
    param.update(dict(title=title))

    if 'visible' not in el.attrib.keys():
        visible=True
    else:
        visible=bool(int(el.get('visible')))
    param.update(dict(visible=visible))

    if 'removable' not in el.attrib.keys():
        removable=False
    else:
        removable=bool(int(el.get('removable')))
    param.update(dict(removable=removable))

    if 'readonly' not in el.attrib.keys():
        readonly=False
    else:
        readonly=bool(int(el.get('readonly')))
    param.update(dict(readonly=readonly))

    if 'show_pb' in el.attrib.keys():
        show_pb = bool(int(el.get('show_pb')))
    else:
        show_pb = False
    param.update(dict(show_pb=show_pb))
    if 'detlist' in el.attrib.keys():
        detlist=eval(el.get('detlist'))
        param.update(dict(detlist=detlist))
    if 'movelist' in el.attrib.keys():
        movelist=eval(el.get('movelist'))
        param.update(dict(movelist=movelist))

    return param

def XML_file_to_parameter(file_name):
    """
        Convert a xml file into pyqtgraph parameter object.

        =============== =========== ================================================
        **Parameters**   **Type**    **Description**

        *file_name*     string      the file name of the XML file to be converted
        =============== =========== ================================================

        Returns
        -------
        params : dictionnary list
            a parameter list of dictionnary to init a parameter

        See Also
        --------
        walk_parameters_to_xml  

        Examples
        --------
        >>> from pyqtgraph.parametertree import Parameter
        >>> import custom_parameter_tree as cpt
        >>>   #Creating test Parameter 
        >>> value=Parameter(name='value',value=10)
        >>> settings=Parameter(name='settings')
        >>> settings.addChild(value)
        <Parameter 'value' at 0x5655f78>
        >>>   #Creating test xml file
        >>> xml_file=cpt.parameter_to_xml_file(settings,"my_xml_file")
        >>> check=cpt.XML_file_to_parameter("my_xml_file.xml")
        >>>   #Verifiy the integrity of the converted parameter
        >>> print(check)
        [{'visible': True, 'type': 'None', 'name': 'value', 'value': '10', 'title': 'value'}]

    """
    tree = ET.parse(file_name)

    root = tree.getroot()
    params=walk_xml_to_parameter(params=[],XML_elt=root)
    return params

def XML_string_to_parameter(xml_string):
    """
        Convert a xml string into pyqtgraph parameter object.

        =============== =========== ================================
        **Parameters**   **Type**    **Description**

        xml_string       string      the xml string to be converted
        =============== =========== ================================

        Returns
        -------
        params: a parameter list of dict to init a parameter

        See Also
        --------
        walk_parameters_to_xml

        Examples
        --------
        >>> import custom_parameter_tree as cpt
        >>> from pyqtgraph.parametertree import Parameter
        >>>    #Creating test parameter
        >>> settings=Parameter(name='settings')
        >>> child=Parameter(name='info', value=10, visible=True)
        >>> settings.addChild(child)
        <Parameter 'info' at 0x84ab558>
        >>>    #Creating test xml string
        >>> xml_string=cpt.parameter_to_xml_string(settings)
        >>>    #Verifiy the integrity of the converted parameter
        >>> converted_parameter=cpt.XML_string_to_parameter(xml_string)
        >>> print(converted_parameter)
        [{'visible': True, 'type': 'None', 'name': 'info', 'value': '10', 'title': 'info'}]

    """
    root = ET.fromstring(xml_string)
    tree=ET.ElementTree(root)

    #tree.write('test.xml')
    params=walk_xml_to_parameter(params=[],XML_elt=root)

    return params

def iter_children(param,childlist=[]):
    """
        | Iterator over all sub children of a given parameters.
        | Returns all childrens names.

        =============== ================================= ====================================
        **Parameters**   **Type**                           **Description**
        *param*          instance of pyqtgraph parameter    the root node to be coursed
        *childlist*      list                               the child list recetion structure
        =============== ================================= ====================================

        Returns
        -------
        childlist : parameter list
            The list of the children from the given node.


        Examples
        --------
        >>> import custom_parameter_tree as cpt
        >>> from pyqtgraph.parametertree import Parameter
        >>>     #Creating the example tree
        >>> settings=Parameter(name='settings')
        >>> child1=Parameter(name='child1', value=10)
        >>> child2=Parameter(name='child2',value=10,visible=True,type='group')
        >>> child2_1=Parameter(name='child2_1', value=10)
        >>> child2_2=Parameter(name='child2_2', value=10)
        >>> child2.addChildren([child2_1,child2_2])
        >>> settings.addChildren([child1,child2])
        >>>     #Get the child list from the param argument
        >>> childlist=cpt.iter_children(settings)
        >>>     #Verify the integrity of result
        >>> print(childlist)
        ['child1', 'child2', 'child2_1', 'child2_2']

    """
    for child in param.children():
        childlist.append(child.name())
        if 'group' in child.type():
            childlist.extend(iter_children(child,[]))
    return childlist

def iter_children_params(param,childlist=[]):
    """

    """
    for child in param.children():
        childlist.append(child)
        if 'group' in child.type():
            childlist.extend(iter_children_params(child,[]))
    return childlist

def get_param_path(param):
    path = [param.name()]
    while param.parent() is not None:
        path.append(param.parent().name())
        param = param.parent()
    return path[::-1]

def get_param_from_name(parent,name):
    for child in parent.children():
        if child.name() == name:
            return child
        if 'group' in child.type():
            ch = get_param_from_name(child, name)
            if ch is not None:
                return ch

class GroupParameterItemCustom(pTypes.GroupParameterItem):
    """
        | Group parameters are used mainly as a generic parent item that holds (and groups!) a set of child parameters. It also provides a simple mechanism for displaying a button or combo that can be used to add new parameters to the group.
        |
        | This customization is made in order to respond to the visible options.
        | Overwrite the optsChanged method from GroupParameterItem class.

    """
    def __init__(self, param, depth):
        pTypes.GroupParameterItem.__init__(self, param, depth)


    def optsChanged(self, param, changed):
        if 'addList' in changed:
            self.updateAddList()
        elif 'visible' in changed:
            self.setHidden(not changed['visible'])


class GroupParameterCustom(pTypes.GroupParameter):
    """
        |
        | Group parameters are used mainly as a generic parent item that holds (and groups!) a set of child parameters.
        |
        | It also provides a simple mechanism for displaying a button or combo that can be used to add new parameters to the group.
        |
        | To enable this, the group  must be initialized with the 'addText' option (the text will be displayed on a button which, when clicked, will cause addNew() to be called).
        |
        | If the 'addList' option is specified as well, then a dropdown-list of addable items will be displayed instead of a button.
        |

        ============== ========================================
        **Attributes**    **Type**
        itemClass         instance of GroupParameterItemCustom
        ============== ========================================
    """
    itemClass = GroupParameterItemCustom
registerParameterType('group', GroupParameterCustom, override=True)



class SpinBoxCustom(SpinBox.SpinBox):
    def __init__(self, parent=None, value=0.0, **kwargs):
        SpinBox.SpinBox.__init__(self,parent=None, value=0.0, **kwargs)
    def setOpts(self, **opts):
        """
            Overriden class to add the field visible in the options.

            =============== =========== ======================
            **Parameters**    **Type**     **Description**
            *opts*            string       the vararg options 
            =============== =========== ======================

            See Also
            --------
            custom_parameter_tree.ItemSelect.setValue
        """
        #print opts
        for k in opts:
            if k == 'bounds':
                self.setMinimum(opts[k][0], update=False)
                self.setMaximum(opts[k][1], update=False)
            elif k == 'min':
                self.setMinimum(opts[k], update=False)
            elif k == 'max':
                self.setMaximum(opts[k], update=False)
            elif k in ['step', 'minStep']:
                self.opts[k] = D(str(opts[k]))
            elif k == 'value':
                pass   ## don't set value until bounds have been set
            elif k == 'visible':
                self.setVisible(opts[k])
            elif k == 'readonly':
                self.setReadOnly(opts[k])
            elif k == 'enabled':
                self.setEnabled(opts[k])

            elif k in self.opts:
                self.opts[k] = opts[k]

            elif k == 'show_pb':
                pass
            elif k == 'subtype':
                pass
            elif k == 'title':
                pass
            elif k == 'type':
                 pass
            else:
                raise TypeError("Invalid keyword argument '%s'." % k)
        if 'value' in opts:
            self.setValue(opts['value'])

        ## If bounds have changed, update value to match
        if 'bounds' in opts and 'value' not in opts:
            self.setValue()

        ## sanity checks:
        if self.opts['int']:
            if 'step' in opts:
                step = opts['step']
                ## not necessary..
                #if int(step) != step:
                    #raise Exception('Integer SpinBox must have integer step size.')
            else:
                self.opts['step'] = int(self.opts['step'])

            if 'minStep' in opts:
                step = opts['minStep']
                if int(step) != step:
                    raise Exception('Integer SpinBox must have integer minStep size.')
            else:
                ms = int(self.opts.get('minStep', 1))
                if ms < 1:
                    ms = 1
                self.opts['minStep'] = ms

        if 'delay' in opts:
            self.proxy.setDelay(opts['delay'])

        self.updateText()

class Pixmap_check(QtWidgets.QWidget):
    """ value of this parameter is a dict with checked, data for the pixmap and optionally path in h5 node
    """
    #valuechanged=pyqtSignal(dict)

    def __init__(self):
        QLocale.setDefault(QLocale(QLocale.English, QLocale.UnitedStates))
        super(Pixmap_check,self).__init__()
        self.path=""
        self.data=None
        self.initUI()

    def initUI(self):
        """
            Init the User Interface.
        """
        self.ver_layout=QtWidgets.QVBoxLayout()
        self.label=QtWidgets.QLabel()
        self.checkbox=QtWidgets.QCheckBox('Show/Hide')
        self.checkbox.setChecked(False)
        self.ver_layout.addWidget(self.label)
        self.ver_layout.addWidget(self.checkbox)
        self.ver_layout.setSpacing(0)
        self.ver_layout.setContentsMargins(0,0,0,0);
        self.setLayout(self.ver_layout)

    def setValue(self,dic):
        if 'data' in dic:
            self.data=QByteArray(dic['data'])
            im=QtGui.QImage.fromData(self.data)
            a=QtGui.QPixmap.fromImage(im)
        else:
            a = dic['pixmap']
        self.label.setPixmap(a)
        self.checkbox.setChecked(dic['checked'])
        self.path=dic['path']
        #self.valuechanged.emit(dic)

    def value(self):
        return dict(pixmap=self.label.pixmap(),checked=self.checkbox.isChecked(),path=self.path)

class QTimeCustom(QtWidgets.QTimeEdit):
    def __init__(self,*args,**kwargs):
        super(QTimeCustom,self).__init__(*args,**kwargs)
        self.minutes_increment=1
        self.timeChanged.connect(self.updateTime)
        
    def setTime(self,time):
        hours=time.hour()
        minutes=time.minute()
        
        minutes=int(np.round(minutes/self.minutes_increment)*self.minutes_increment)
        if minutes==60:
            minutes=0
            hours+=1
            
        time.setHMS(hours,minutes,0)
        
        return super(QTimeCustom,self).setTime(time)
        
            
    def setMinuteIncrement(self,minutes_increment):
        self.minutes_increment=minutes_increment
        self.updateTime(self.time())
        
    @pyqtSlot(QTime)
    def updateTime(self,time):
        self.setTime(time)


class SliderParameterItem(pTypes.WidgetParameterItem):

    def __init__(self, param, depth):
        pTypes.WidgetParameterItem.__init__(self, param, depth)
        self.hideWidget = False
        self.subItem = QtWidgets.QTreeWidgetItem()
        self.addChild(self.subItem)

    def treeWidgetChanged(self):
        ## TODO: fix so that superclass method can be called
        ## (WidgetParameter should just natively support this style)
        # WidgetParameterItem.treeWidgetChanged(self)
        self.treeWidget().setFirstItemColumnSpanned(self.subItem, True)
        self.treeWidget().setItemWidget(self.subItem, 0, self.w)

        # for now, these are copied from ParameterItem.treeWidgetChanged
        self.setHidden(not self.param.opts.get('visible', True))
        self.setExpanded(self.param.opts.get('expanded', True))

    def limitsChanged(self, param, limits):
        """Called when the parameter's limits have changed"""
        ParameterItem.limitsChanged(self, param, limits)
        t = self.param.opts['type']
        self.w.setOpts(bounds=limits)

    def makeWidget(self):
        """
            Make an initialized file_browser object with parameter options dictionnary ('readonly' key)0

            Returns
            -------
            w : filebrowser
                The initialized file browser.

            See Also
            --------
            file_browser
        """
        opts = self.param.opts

        defs = {
            'value': 0, 'min': 0., 'max': 1000.,
            'step': 1.0, 'dec': False,
            'siPrefix': False, 'suffix': '', 'decimals': 12,
        }

        for k in defs:
            if k in opts:
                defs[k] = opts[k]
        if 'limits' in opts:
            defs['bounds'] = opts['limits']
        if 'subtype' not in opts:
            opts['subtype'] = 'linear'
        self.w = SliderSpinBox(bounds=(0., 1000.), subtype=opts['subtype'])
        self.w.setOpts(**defs)

        self.w.sigChanged = self.w.spinbox.sigValueChanged
        self.w.sigChanging = self.w.spinbox.sigValueChanging

        return self.w


class SliderParameter(Parameter):
    """
        Editable string; displayed as large text box in the tree.

        See Also
        --------
        file_browserParameterItem
    """
    itemClass = SliderParameterItem




registerParameterType('slide', SliderParameter, override=True)

class SliderSpinBox(QtWidgets.QWidget):

    def __init__(self,*args,**kwargs):
        QLocale.setDefault(QLocale(QLocale.English, QLocale.UnitedStates))
        super(SliderSpinBox,self).__init__()
        self.subtype = kwargs['subtype']
        self.initUI(*args,**kwargs)
        
    @property
    def opts(self):
        return self.spinbox.opts

    @opts.setter
    def opts(self,**opts):
        self.setOpts(**opts)

    def setOpts(self, **opts):
        self.spinbox.setOpts(**opts)
        if 'visible' in opts:
            self.slider.setVisible(opts['visible'])


    def initUI(self,*args,**kwargs):
        """
            Init the User Interface.
        """
        self.hor_layout=QtWidgets.QVBoxLayout()
        self.slider=QtWidgets.QSlider(Qt.Horizontal)
        self.slider.setMinimum(0)
        self.slider.setMaximum(100)

        self.spinbox=SpinBoxCustom(parent=None, value=1, **kwargs)

        self.hor_layout.addWidget(self.slider)
        self.hor_layout.addWidget(self.spinbox)
        self.hor_layout.setSpacing(0)
        self.hor_layout.setContentsMargins(0,0,0,0);
        self.setLayout(self.hor_layout)

        self.slider.valueChanged.connect(self.update_spinbox)
        self.spinbox.valueChanged.connect(self.update_slide)


    def update_spinbox(self,val):
        """
        val is a percentage [0-100] used in order to set the spinbox value between its min and max
        """
        min_val=float(self.opts['bounds'][0])
        max_val=float(self.opts['bounds'][1])
        if self.subtype == 'log':
            val_out = scroll_log(val, min_val, max_val)
        else:
            val_out = scroll_linear(val, min_val, max_val)
        try:
            self.slider.valueChanged.disconnect(self.update_spinbox)
            self.spinbox.valueChanged.disconnect(self.update_slide)
        except:
            pass
        self.spinbox.setValue(val_out)

        self.slider.valueChanged.connect(self.update_spinbox)
        self.spinbox.valueChanged.connect(self.update_slide)

    def update_slide(self,val):
        """
        val is the spinbox value between its min and max
        """
        min_val=float(self.opts['bounds'][0])
        max_val=float(self.opts['bounds'][1])


        try:
            self.slider.valueChanged.disconnect(self.update_spinbox)
            self.spinbox.valueChanged.disconnect(self.update_slide)
        except:
            pass
        self.slider.setValue(int((val-min_val)/(max_val-min_val)*100))
        self.slider.valueChanged.connect(self.update_spinbox)
        self.spinbox.valueChanged.connect(self.update_slide)

    def setValue(self,val):
        self.spinbox.setValue(val)

    def value(self):
        return self.spinbox.value()

class WidgetParameterItemcustom(pTypes.WidgetParameterItem):
    """
        This is a subclass of widget parameteritem in order to deal with the visiblily of the spinbox when parameter visibility os toggled.
    """
    def __init__(self, param, depth):
        pTypes.WidgetParameterItem.__init__(self, param, depth)

        if 'enabled' in self.param.opts:
            self.displayLabel.setEnabled(self.param.opts['enabled'])

    def makeWidget(self):
        """
            | Return a single widget that should be placed in the second tree column.
            | The widget must be given three attributes:

            ==========  ============================================================
            sigChanged  a signal that is emitted when the widget's value is changed
            value       a function that returns the value
            setValue    a function that sets the value
            ==========  ============================================================

            | This is a good function to override in subclasses.

            See Also
            --------
            SpinBoxCustom, custom_parameter_tree.setOpts
        """
        opts = self.param.opts
        t = opts['type']
        if t in ('int', 'float'):
            defs = {
                'value': 0, 'min': None, 'max': None,
                'step': 1.0, 'dec': False,
                'siPrefix': False, 'suffix': '', 'decimals': 12,
            }
            if t == 'int':
                defs['int'] = True
                defs['minStep'] = 1.0
            for k in defs:
                if k in opts:
                    defs[k] = opts[k]
            if 'limits' in opts:
                defs['bounds'] = opts['limits']
            w = SpinBoxCustom()
            w.setOpts(**defs)
            w.sigChanged = w.sigValueChanged
            w.sigChanging = w.sigValueChanging
        elif t == 'bool':
            w = QtWidgets.QCheckBox()
            w.sigChanged = w.toggled
            w.value = w.isChecked
            w.setValue = w.setChecked
            w.setEnabled(not opts.get('readonly', False))
            self.hideWidget = False
        elif t == 'bool_push':
            w = QtWidgets.QPushButton()
            if 'title' in opts:
                w.setText(opts['title'])
            else:
                w.setText(opts['name'])
            w.setMaximumWidth(50)
            w.setCheckable(True)
            w.sigChanged = w.toggled
            w.value = w.isChecked
            w.setValue = w.setChecked
            w.setEnabled(not opts.get('readonly', False))
            self.hideWidget = False
        elif t == 'str':
            w = QtWidgets.QLineEdit()
            w.sigChanged = w.editingFinished
            w.value = lambda: str(w.text())
            w.setValue = lambda v: w.setText(str(v))
            w.sigChanging = w.textChanged
        elif t == 'color':
            w = ColorButton.ColorButton()
            w.sigChanged = w.sigColorChanged
            w.sigChanging = w.sigColorChanging
            w.value = w.color
            w.setValue = w.setColor
            self.hideWidget = False
            w.setFlat(True)
            w.setEnabled(not opts.get('readonly', False))
        elif t == 'colormap':
            from pyqtgraph.widgets.GradientWidget import GradientWidget ## need this here to avoid import loop
            w = GradientWidget(orientation='bottom')
            w.sigChanged = w.sigGradientChangeFinished
            w.sigChanging = w.sigGradientChanged
            w.value = w.colorMap
            w.setValue = w.setColorMap
            self.hideWidget = False
        elif t== 'date_time':
            w=QtWidgets.QDateTimeEdit(QDateTime(QDate.currentDate(),QTime.currentTime()))
            w.setCalendarPopup(True)
            w.setDisplayFormat('dd/MM/yyyy hh:mm')
            w.sigChanged=w.dateTimeChanged
            w.value = w.dateTime
            w.setValue= w.setDateTime
        elif t== 'date':
            w=QtWidgets.QDateEdit(QDate(QDate.currentDate()))
            w.setCalendarPopup(True)
            w.setDisplayFormat('dd/MM/yyyy')
            w.sigChanged=w.dateChanged
            w.value = w.date
            w.setValue= w.setDate
            
        elif t== 'time':
            w=QTimeCustom(QTime(QTime.currentTime()))
            if 'minutes_increment' in opts:
                w.setMinuteIncrement(opts['minutes_increment'])
            w.setDisplayFormat('hh:mm')
            w.sigChanged=w.timeChanged
            w.value = w.time
            w.setValue= w.setTime

        elif t=='led':
            w=QLED()
            w.clickable=False
            w.set_as_false()
            w.sigChanged=w.value_changed
            w.value = w.get_state
            w.setValue= w.set_as
        elif t=='pixmap':
            w=QtWidgets.QLabel()
            w.sigChanged=None
            w.value = w.pixmap
            w.setValue= w.setPixmap
        elif t=='pixmap_check':
            w=Pixmap_check()
            w.sigChanged=w.checkbox.toggled
            w.value = w.value
            w.setValue= w.setValue
        # elif t=='slide':
        #
        #     defs = {
        #         'value': 0, 'min': 0., 'max': 1000.,
        #         'step': 1.0, 'dec': False,
        #         'siPrefix': False, 'suffix': '', 'decimals': 12,
        #     }
        #     for k in defs:
        #         if k in opts:
        #             defs[k] = opts[k]
        #     if 'limits' in opts:
        #         defs['bounds'] = opts['limits']
        #     if 'subtype' not in opts:
        #         opts['subtype'] = 'linear'
        #     w = SliderSpinBox(bounds=(0.,1000.), subtype=opts['subtype'])
        #     w.setOpts(**defs)
        #
        #     w.sigChanged = w.spinbox.sigValueChanged
        #     w.sigChanging = w.spinbox.sigValueChanging

        else:
            raise Exception("Unknown type '%s'" % str(t))
        return w

    def showEditor(self):
        """
            Show the widget attribute.
        """
        self.widget.show()
        self.displayLabel.hide()
        self.widget.setFocus(Qt.OtherFocusReason)
        if isinstance(self.widget, SpinBox.SpinBox):
            self.widget.selectNumber()  # select the numerical portion of the text for quick editing

    def hideEditor(self):
        """
            Hide the widget attribute.
        """
        if not (self.param.opts['type']=='led' or self.param.opts['type']=='pixmap' or self.param.opts['type']=='pixmap_check'):
            self.widget.hide()
            self.displayLabel.show()

    def optsChanged(self, param, opts):
        """
            | Called when any options are changed that are not name, value, default, or limits.
            |
            | If widget is a SpinBox, pass options straight through.
            | So that only the display label is shown when visible option is toggled.

            =============== ================================== ==============================
            **Parameters**    **Type**                           **Description**
            *param*           instance of pyqtgraph parameter    the parameter to check
            *opts*            string list                        the associated options list
            =============== ================================== ==============================

            See Also
            --------
            optsChanged
        """
        #print "opts changed:", opts
        ParameterItem.optsChanged(self, param, opts)

        if 'readonly' in opts:
            self.updateDefaultBtn()
            if isinstance(self.widget, (QtWidgets.QCheckBox,ColorButton.ColorButton)):
                self.widget.setEnabled(not opts['readonly'])

        if 'minutes_increment' in opts:
            self.widget.setMinuteIncrement(opts['minutes_increment'])

        ## If widget is a SpinBox, pass options straight through
        if isinstance(self.widget, SpinBoxCustom):
            if 'visible' in opts:
                opts.pop('visible')
                self.widget.hide() # so that only the display label is shown when visible option is toggled
            if 'units' in opts and 'suffix' not in opts:
                opts['suffix'] = opts['units']
            self.widget.setOpts(**opts)
            self.updateDisplayLabel()



    def valueChanged(self, param, val, force=False):
        ## called when the parameter's value has changed
        ParameterItem.valueChanged(self, param, val)
        if self.widget.sigChanged is not None:
            self.widget.sigChanged.disconnect(self.widgetValueChanged)

        try:
            if force or val != self.widget.value():
                self.widget.setValue(val)
            self.updateDisplayLabel(val)  ## always make sure label is updated, even if values match!
        finally:
            if self.widget.sigChanged is not None:
                self.widget.sigChanged.connect(self.widgetValueChanged)
        self.updateDefaultBtn()

class SimpleParameterCustom(pTypes.SimpleParameter):
    itemClass = WidgetParameterItemcustom

    def __init__(self, *args, **kargs):
        pTypes.SimpleParameter.__init__(self, *args, **kargs)

    # def _interpretValue(self, v):
    #     fn = {
    #         'int': int,
    #         'float': float,
    #         'bool': bool,
    #         'str': str,
    #         'color': self._interpColor,
    #         'colormap': self._interpColormap,
    #         'date_time': QDateTime,
    #         'date': QDate,
    #         'time': QTime,
    #         'led': qled,
    #         'pixmap': QtWidgets.QLabel,
    #         'pixmap_check': Pixmap_check,
    #         'slide': float
    #     }[self.opts['type']]
    #     return fn(v)


registerParameterType('int',SimpleParameterCustom, override=True)
registerParameterType('float', SimpleParameterCustom , override=True)
registerParameterType('bool',SimpleParameterCustom, override=True)
registerParameterType('bool_push',SimpleParameterCustom, override=True)
registerParameterType('date_time', SimpleParameterCustom , override=True)
registerParameterType('date', SimpleParameterCustom , override=True)
registerParameterType('time', SimpleParameterCustom , override=True)
registerParameterType('led', SimpleParameterCustom , override=True)
registerParameterType('pixmap', SimpleParameterCustom , override=True)
registerParameterType('pixmap_check', SimpleParameterCustom , override=True)
# registerParameterType('slide', SimpleParameterCustom , override=True)

class ListParameterItem_custom(pTypes.ListParameterItem):
    """
        WidgetParameterItem subclass providing comboBox that lets the user select from a list of options.

    """
    def __init__(self, param, depth):
        super(ListParameterItem_custom,self).__init__(param, depth)


    def makeWidget(self):
        """
            Make a widget from self parameter options, connected to the buttonClicked function.

            Returns
            -------
            w:widget
                the initialized widget

            See Also
            --------
            buttonClicked, limitsChanged, custom_parameter_tree.ItemSelect.setValue
        """
        opts = self.param.opts
        t = opts['type']
        w = Combo_pb()
        w.add_pb.clicked.connect(self.buttonClicked)
        w.setMaximumHeight(20)  ## set to match height of spin box and line edit
        if 'show_pb' in opts:
            w.add_pb.setVisible(opts['show_pb'])
        else:
            w.add_pb.setVisible(False)
        w.sigChanged = w.combo.currentIndexChanged
        w.value = self.value
        w.setValue = self.setValue
        self.widget = w  ## needs to be set before limits are changed
        self.limitsChanged(self.param, self.param.opts['limits'])
        if len(self.forward) > 0:
            self.setValue(self.param.value())
        return w

    def value(self):
        key = str(self.widget.combo.currentText())
        return self.forward.get(key, None)

    def setValue(self, val):
        self.targetValue = val
        if val not in self.reverse[0]:
            self.widget.combo.setCurrentIndex(0)
        else:
            key = self.reverse[1][self.reverse[0].index(val)]
            ind = self.widget.combo.findText(key)
            self.widget.combo.setCurrentIndex(ind)

    def limitsChanged(self, param, limits):
        """
            Set up forward / reverse mappings for {name:value} limits dictionnary.

            =============== ================================== ========================================
            **Parameters**    **Type**                          **Description**
            *param*           instance of pyqtgraph parameter    Not used
            *limits*          dictionnary                        the limits dictionnary to be mapped
            =============== ================================== ========================================

        """

        if len(limits) == 0:
            limits = ['']  ## Can never have an empty list--there is always at least a singhe blank item.

        self.forward, self.reverse = ListParameter_custom.mapping(limits)
        try:
            self.widget.blockSignals(True)
            val = self.targetValue  #asUnicode(self.widget.currentText())

            self.widget.combo.clear()
            for k in self.forward:
                self.widget.combo.addItem(k)
                if k == val:
                    self.widget.combo.setCurrentIndex(self.widget.count()-1)
                    self.updateDisplayLabel()
        finally:
            self.widget.blockSignals(False)

    def buttonClicked(self):
        """
            |
            | Append the self limits attributes an added parameter with string value.
            | Update parameter and call the limitschanged method to map the added parameter.

            See Also
            --------
            limitsChanged, custom_parameter_tree.ItemSelect.setValue
        """
        if type(self.param.opts['limits']) == list:
            text,ok = QtWidgets.QInputDialog.getText(None,"Enter a value to add to the parameter",
                                             "String value:", QtWidgets.QLineEdit.Normal);
            if ok and not (text==""):
                self.param.opts['limits'].append(text)
                self.limitsChanged(self.param,self.param.opts['limits'])
                self.param.setValue(text)

    def optsChanged(self, param, opts):
        """
            Called when any options are changed that are not name, value, default, or limits.

            =============== ================================== =======================================
            **Parameters**    **Type**                           **Description**
            *param*           instance of pyqtgraph parameter    The parameter to be checked
            *opts*            string list                        The option dictionnary to be checked
            =============== ================================== =======================================

            See Also
            --------
            optsChanged
        """
        #print "opts changed:", opts
        ParameterItem.optsChanged(self, param, opts)

        if 'show_pb' in opts:
            self.widget.add_pb.setVisible(opts['show_pb'])
        if 'enabled' in opts:
            self.widget.setEnabled(opts['enabled'])


class ListParameter_custom(pTypes.ListParameter):
    """
        =============== =======================================
        **Attributes**    **Type**
        *itemClass*       instance of ListParameterItem_custom
        *sigActivated*    instance of pyqt Signal
        =============== =======================================
    """
    itemClass = ListParameterItem_custom
    sigActivated = pyqtSignal(object)
    def __init__(self, **opts):
        super(ListParameter_custom,self).__init__( **opts)


    def activate(self):
        """
            Emit the Activated signal.
        """
        self.sigActivated.emit(self)
        self.emitStateChanged('activated', None)


registerParameterType('list', ListParameter_custom, override=True)





class Combo_pb(QtWidgets.QWidget):

    def __init__(self,items=[]):
        QLocale.setDefault(QLocale(QLocale.English, QLocale.UnitedStates))
        super(Combo_pb,self).__init__()
        self.items=items
        self.initUI()
        self.count=self.combo.count

    def initUI(self):
        """
            Init the User Interface.
        """


        self.hor_layout=QtWidgets.QHBoxLayout()
        self.combo=QtWidgets.QComboBox()
        self.combo.addItems(self.items)
        self.add_pb=QtWidgets.QPushButton()
        self.add_pb.setText("")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/Add2.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.add_pb.setIcon(icon3)
        self.hor_layout.addWidget(self.combo)
        self.hor_layout.addWidget(self.add_pb)
        self.hor_layout.setSpacing(0)
        self.hor_layout.setContentsMargins(0,0,0,0);
        self.add_pb.setMaximumWidth(25)
        self.setLayout(self.hor_layout)


class TableParameterItem(pTypes.WidgetParameterItem):

    def __init__(self, param, depth):
        pTypes.WidgetParameterItem.__init__(self, param, depth)
        self.hideWidget = False
        self.subItem = QtGui.QTreeWidgetItem()
        self.addChild(self.subItem)

    def treeWidgetChanged(self):
        """
            Check for changement in the Widget tree.
        """
        ## TODO: fix so that superclass method can be called
        ## (WidgetParameter should just natively support this style)
        #WidgetParameterItem.treeWidgetChanged(self)
        self.treeWidget().setFirstItemColumnSpanned(self.subItem, True)
        self.treeWidget().setItemWidget(self.subItem, 0, self.widget)

        # for now, these are copied from ParameterItem.treeWidgetChanged
        self.setHidden(not self.param.opts.get('visible', True))
        self.setExpanded(self.param.opts.get('expanded', True))

    def makeWidget(self):
        """
            Make and initialize an instance of Table_custom.

            Returns
            -------
            table : instance of Table_custom.
                The initialized table.

            See Also
            --------
            Table_custom
        """
        w = Table_custom()
        w.setColumnCount(2)
        if 'header' in self.param.opts.keys():
            w.setHorizontalHeaderLabels(self.param.opts['header'])
        w.setMaximumHeight(200)
        #self.table.setReadOnly(self.param.opts.get('readonly', False))
        w.value = w.get_table_value
        w.setValue = w.set_table_value
        w.sigChanged = w.itemChanged
        return w



class Table_custom(QtWidgets.QTableWidget):
    """
        ============== ===========================
        *Attributes**    **Type**
        *valuechanged*   instance of pyqt Signal
        *QtWidgets*      instance of QTableWidget
        ============== ===========================
    """
    valuechanged=pyqtSignal(OrderedDict)
    def __init__(self):
        QtWidgets.QTableWidget.__init__(self)



    def get_table_value(self):
        """
            Get the contents of the self coursed table.

            Returns
            -------
            data : ordered dictionnary
                The getted values dictionnary.
        """
        data = OrderedDict([])
        for ind in range(self.rowCount()):
            item0 = self.item(ind,0)
            item1 = self.item(ind,1)
            if item0 is not None and item1 is not None:
                try:
                    data[item0.text()] = float(item1.text())
                except:
                    data[item0.text()] = item1.text()
        return data


    def set_table_value(self,data_dict):
        """
            Set the data values dictionnary to the custom table.

            =============== ====================== ================================================
            **Parameters**    **Type**               **Description**
            *data_dict*       ordered dictionnary    the contents to be stored in the custom table
            =============== ====================== ================================================
        """
        try:
            self.setRowCount(len(data_dict))
            self.setColumnCount(2)
            for ind, (key, value) in enumerate(data_dict.items()):
                item0 = QtWidgets.QTableWidgetItem(key)
                item0.setFlags(item0.flags() ^ Qt.ItemIsEditable)
                if isinstance(value,float):
                    item1 = QtWidgets.QTableWidgetItem('{:.6e}'.format(value))
                else:
                    item1 = QtWidgets.QTableWidgetItem(str(value))
                item1.setFlags(item1.flags() ^ Qt.ItemIsEditable)
                self.setItem(ind, 0, item0)
                self.setItem(ind, 1, item1)
            #self.valuechanged.emit(data_dict)

        except Exception as e:
            pass

class TableParameter(Parameter):
    """
        =============== =================================
        **Attributes**    **Type**
        *itemClass*       instance of TableParameterItem
        *Parameter*       instance of pyqtgraph parameter
        =============== =================================
    """
    itemClass = TableParameterItem
    """Editable string; displayed as large text box in the tree."""
    # def __init(self):
    #     super(TableParameter,self).__init__()

    def setValue(self,value):
        self.opts['value'] = value
        self.sigValueChanged.emit(self, value)

registerParameterType('table', TableParameter, override=True)




class ItemSelectParameterItem(pTypes.WidgetParameterItem):

    def __init__(self, param, depth):
        pTypes.WidgetParameterItem.__init__(self, param, depth)
        self.hideWidget = False
        self.subItem = QtWidgets.QTreeWidgetItem()
        self.addChild(self.subItem)

    def treeWidgetChanged(self):
        """

        """
        ## TODO: fix so that superclass method can be called
        ## (WidgetParameter should just natively support this style)
        #WidgetParameterItem.treeWidgetChanged(self)
        self.treeWidget().setFirstItemColumnSpanned(self.subItem, True)
        self.treeWidget().setItemWidget(self.subItem, 0, self.widget)

        # for now, these are copied from ParameterItem.treeWidgetChanged
        self.setHidden(not self.param.opts.get('visible', True))
        self.setExpanded(self.param.opts.get('expanded', True))

    def makeWidget(self):
        """
            | Make and initialize an instance of ItemSelect_pb with itemselect value.
            | Connect the created object with the buttonClicked function.

        """
        opts = self.param.opts
        w = ItemSelect_pb()
        w.itemselect.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        w.itemselect.setMaximumHeight(70)
        #w.setReadOnly(self.param.opts.get('readonly', False))
        if 'show_pb' in opts:
            w.add_pb.setVisible(opts['show_pb'])
        else:
            w.add_pb.setVisible(False)
        w.value = w.itemselect.get_value
        w.setValue = w.itemselect.set_value
        w.sigChanged = w.itemselect.itemSelectionChanged
        w.add_pb.clicked.connect(self.buttonClicked)
        return w

    def buttonClicked(self):
        """
           Append to the param attribute the dictionnary obtained from the QtWidget add parameter procedure.

           See Also
           --------
           custom_parameter_tree.value, custom_parameter_tree.ItemSelect.setValue
        """

        text,ok = QtWidgets.QInputDialog.getText(None,"Enter a value to add to the parameter",
                                            "String value:", QtWidgets.QLineEdit.Normal);
        if ok and not (text==""):
            all=self.param.value()['all_items']
            all.append(text)
            sel=self.param.value()['selected']
            sel.append(text)
            val=dict(all_items=all,selected=sel)
            self.param.setValue(val)
            self.param.sigValueChanged.emit(self.param, val)

    def optsChanged(self, param, opts):
        """
            Called when any options are changed that are not name, value, default, or limits.

            See Also
            --------
            optsChanged
        """
        #print "opts changed:", opts
        ParameterItem.optsChanged(self, param, opts)

        if 'show_pb' in opts:
            self.widget.add_pb.setVisible(opts['show_pb'])


class ItemSelect_pb(QtWidgets.QWidget):
    def __init__(self):
        QLocale.setDefault(QLocale(QLocale.English, QLocale.UnitedStates))
        super(ItemSelect_pb,self).__init__()
        self.initUI()

    def initUI(self):
        self.hor_layout=QtWidgets.QHBoxLayout()
        self.itemselect=ItemSelect()
        self.add_pb=QtWidgets.QPushButton()
        self.add_pb.setText("")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/Add2.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.add_pb.setIcon(icon3)
        self.hor_layout.addWidget(self.itemselect)
        self.hor_layout.addWidget(self.add_pb)
        self.hor_layout.setSpacing(0)

        self.setLayout(self.hor_layout)


class ItemSelect(QtWidgets.QListWidget):
    def __init__(self):
        QtWidgets.QListWidget.__init__(self)

    def get_value(self):
        """
            Get the dictionnary of values contained in the QtWidget attribute.

            Returns
            -------
            dictionnary
                The dictionnary of all_items compared to the slelectedItems.
        """
        selitems=[item.text() for item in self.selectedItems()]
        allitems=[item.text() for item in self.all_items()]
        return dict(all_items=allitems,selected=selitems)

    def all_items(self):
        """
            Get the all_items list from the self QtWidget attribute.

            Returns
            -------
            list
                The item list.
        """
        return [self.item(ind) for ind in range(self.count())]

    def set_value(self,values):
        """
            Set values to the all_items attributes filtering values by the 'selected' key.

            =============== ============== =======================================
            **Parameters**    **Type**       **Description**
            *values*          dictionnary    the values dictionnary to be setted.
            =============== ============== =======================================
        """
        allitems=[item.text() for item in self.all_items()]
        if allitems !=values['all_items']:
            self.clear()
            self.addItems(values['all_items'])
            QtWidgets.QApplication.processEvents()
        for item in self.all_items():
            if item.text() in values['selected']:
                item.setSelected(True)

class ItemSelectParameter(Parameter):
    """
        Editable string; displayed as large text box in the tree.

        =============== ======================================
        **Attributes**    **Type**
        *itemClass*       instance of ItemSelectParameterItem
        *sigActivated*    instance of pyqt Signal
        =============== ======================================
    """
    itemClass = ItemSelectParameterItem
    sigActivated = pyqtSignal(object)

    def activate(self):
        """
            Activate the "Activated" signal attribute0
        """
        self.sigActivated.emit(self)
        self.emitStateChanged('activated', None)

registerParameterType('itemselect', ItemSelectParameter, override=True)

class file_browserParameterItem(pTypes.WidgetParameterItem):

    def __init__(self, param, depth):
        self.filetype=False
        pTypes.WidgetParameterItem.__init__(self, param, depth)
        self.hideWidget = False
        self.subItem = QtWidgets.QTreeWidgetItem()
        self.addChild(self.subItem)
        

    def treeWidgetChanged(self):
        ## TODO: fix so that superclass method can be called
        ## (WidgetParameter should just natively support this style)
        #WidgetParameterItem.treeWidgetChanged(self)
        self.treeWidget().setFirstItemColumnSpanned(self.subItem, True)
        self.treeWidget().setItemWidget(self.subItem, 0, self.w)

        # for now, these are copied from ParameterItem.treeWidgetChanged
        self.setHidden(not self.param.opts.get('visible', True))
        self.setExpanded(self.param.opts.get('expanded', True))

    def makeWidget(self):
        """
            Make an initialized file_browser object with parameter options dictionnary ('readonly' key)0

            Returns
            -------
            w : filebrowser
                The initialized file browser.

            See Also
            --------
            file_browser
        """
        if 'filetype' in self.param.opts:
            self.filetype = self.param.opts['filetype']
        else:
            self.filetype = True

        self.w = file_browser(self.param.value(),file_type=self.filetype)
        #self.file_browser.setMaximumHeight(100)
        self.w.base_path_edit.setReadOnly(self.param.opts['readonly'])
        self.w.value = self.w.get_value
        self.w.setValue = self.w.set_path
        self.w.sigChanged = self.w.value_changed
        return self.w

class file_browser(QtWidgets.QWidget):
    """
        ================ =========================
        **Attributes**    **Type**
        *value_changed*   instance of pyqt Signal
        *path*            string
        ================ =========================

        See Also
        --------
        browse_path
    """
    value_changed=pyqtSignal(str)
    def __init__(self,init_path='D:/Data',file_type=False):
        QLocale.setDefault(QLocale(QLocale.English, QLocale.UnitedStates))
        super(file_browser,self).__init__()
        self.filetype=file_type
        self.path=init_path
        self.initUI()

        self.base_path_browse_pb.clicked.connect(self.browse_path)


    def browse_path(self):
        """
            Browse the path attribute if exist.

            See Also
            --------
            set_path
        """
        if self.filetype is True:
            folder_name = QtWidgets.QFileDialog.getOpenFileName(None,'Choose File',os.path.split(self.path)[0])[0]
        elif self.filetype is False:
            folder_name = QtWidgets.QFileDialog.getExistingDirectory(None,'Choose Folder',self.path)

        elif self.filetype == "save":
            folder_name = QtWidgets.QFileDialog.getSaveFileName(None,'Enter a Filename', os.path.split(self.path)[0])[0]

        if not( not(folder_name)): #execute if the user didn't cancel the file selection
             self.set_path(folder_name)
             self.value_changed.emit(folder_name)

    def set_path(self,path_file):
        """
            Set the base path attribute with the given path_file.

            =============== =========== ===========================
            **Parameters**    **Type**    **Description**
            *path_file*       string      the pathname of the file
            =============== =========== ===========================
        """
        self.base_path_edit.setPlainText(path_file)
        self.path = path_file

    def get_value(self):
        """
            Get the value of the base_path_edit attribute.

            Returns
            -------
            string
                the path name
        """
        return self.base_path_edit.toPlainText()


    def initUI(self):
        """
            Init the User Interface.
        """

        self.hor_layout=QtWidgets.QHBoxLayout()
        self.base_path_edit=QtWidgets.QPlainTextEdit(self.path)
        self.base_path_edit.setMaximumHeight(50)
        self.base_path_browse_pb=QtWidgets.QPushButton()
        self.base_path_browse_pb.setText("")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/Browse_Dir_Path.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.base_path_browse_pb.setIcon(icon3)
        self.hor_layout.addWidget(self.base_path_edit)

        verlayout=QtWidgets.QVBoxLayout()
        verlayout.addWidget(self.base_path_browse_pb)
        verlayout.addStretch()
        self.hor_layout.addLayout(verlayout)
        self.hor_layout.setSpacing(0)
        self.setLayout(self.hor_layout)



class file_browserParameter(Parameter):
    """
        Editable string; displayed as large text box in the tree.
        
        See Also
        --------
        file_browserParameterItem
    """
    itemClass = file_browserParameterItem

registerParameterType('browsepath', file_browserParameter, override=True)

class Plain_text_pbParameterItem(pTypes.WidgetParameterItem):

    def __init__(self, param, depth):
        pTypes.WidgetParameterItem.__init__(self, param, depth)
        self.hideWidget = False
        self.subItem = QtWidgets.QTreeWidgetItem()
        self.addChild(self.subItem)

    def treeWidgetChanged(self):
        ## TODO: fix so that superclass method can be called
        ## (WidgetParameter should just natively support this style)
        #WidgetParameterItem.treeWidgetChanged(self)
        self.treeWidget().setFirstItemColumnSpanned(self.subItem, True)
        self.treeWidget().setItemWidget(self.subItem, 0, self.w)

        # for now, these are copied from ParameterItem.treeWidgetChanged
        self.setHidden(not self.param.opts.get('visible', True))
        self.setExpanded(self.param.opts.get('expanded', True))

    def makeWidget(self):
        """
            Make and initialize an instance of Plain_text_pb object from parameter options dictionnary (using 'readonly' key).

            Returns
            -------
            Plain_text_pb object
                The initialized object.

            See Also
            --------
            Plain_text_pb, buttonClicked
        """
        self.w = Plain_text_pb()
        self.w.text_edit.setReadOnly(self.param.opts.get('readonly', False))
        self.w.value = self.w.get_value
        self.w.setValue = self.w.set_value
        self.w.sigChanged = self.w.value_changed
        self.w.add_pb.clicked.connect(self.buttonClicked)
        return self.w

    def buttonClicked(self):
        """
            Activate the parameter attribute.

            See Also
            --------
            custom_parameter_tree.ItemSelectParameter.activate
        """
        self.param.activate()

class Plain_text_pb(QtWidgets.QWidget):
    """
        ================ ========================
        **Attributes**    **Type**
        *value_changed*   instance of pyqt Signal
        ================ ========================

        See Also
        --------
        initUI, emitsignal
    """
    value_changed=pyqtSignal(str)
    def __init__(self):
        QLocale.setDefault(QLocale(QLocale.English, QLocale.UnitedStates))
        super(Plain_text_pb,self).__init__()

        self.initUI()
        self.text_edit.textChanged.connect(self.emitsignal)

    def emitsignal(self):
        """
            Emit the value changed signal from the text_edit attribute.
        """
        text=self.text_edit.toPlainText()
        self.value_changed.emit(text)

    def set_value(self,txt):
        """
            Set the value of the text_edit attribute.

            =============== =========== ================================
            **Parameters**    **Type**    **Description**
            *txt*             string      the string value to be setted
            =============== =========== ================================
        """
        self.text_edit.setPlainText(txt)

    def get_value(self):
        """
            Get the value of the text_edit attribute.

            Returns
            -------
            string
                The string value of text_edit.
        """
        return self.text_edit.toPlainText()


    def initUI(self):
        """
            Init the User Interface.
        """

        self.hor_layout=QtWidgets.QHBoxLayout()
        self.text_edit=QtWidgets.QPlainTextEdit()
        self.text_edit.setReadOnly(True)
        self.text_edit.setMaximumHeight(50)

        self.add_pb=QtWidgets.QPushButton()
        self.add_pb.setText("")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/Add2.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.add_pb.setIcon(icon3)
        self.hor_layout.addWidget(self.text_edit)


        verlayout=QtWidgets.QVBoxLayout()
        verlayout.addWidget(self.add_pb)
        verlayout.addStretch()
        self.hor_layout.addLayout(verlayout)
        self.hor_layout.setSpacing(0)
        self.setLayout(self.hor_layout)

class Plain_text_pbParameter(Parameter):
    """Editable string; displayed as large text box in the tree."""
    itemClass = Plain_text_pbParameterItem
    sigActivated = pyqtSignal(object)

    def activate(self):
        """
            Send the Activated signal.
        """
        self.sigActivated.emit(self)
        self.emitStateChanged('activated', None)
registerParameterType('text_pb', Plain_text_pbParameter, override=True)


class TextParameterItemCustom(pTypes.TextParameterItem):
    def __init__(self, param, depth):
        super(TextParameterItemCustom, self).__init__(param, depth)

        self.textBox.setMaximumHeight(50)



class TextParameter(Parameter):
    """Editable string; displayed as large text box in the tree."""
    itemClass = TextParameterItemCustom
registerParameterType('text', TextParameter, override=True)



if __name__ == '__main__':

    app = QtWidgets.QApplication(sys.argv);
    ex=QTimeCustom()
    ex.setMinuteIncrement(30)
    ex.show()
    sys.exit(app.exec_())



from PyQt5 import QtWidgets
from PyQt5.QtCore import QDateTime, QTime, QDate, QObject, pyqtSignal
from pyqtgraph.parametertree import Parameter, ParameterTree
import pymodaq.daq_utils.custom_parameter_tree as custom_tree
from pymodaq.daq_utils.h5browser import H5Browser
from pymodaq.daq_utils import daq_utils as utils
from pymodaq.version import get_version
import datetime
import os
import tables
import numpy as np
from pathlib import Path
import copy

version = '0.0.1'
save_types = ['scan', 'detector', 'custom']
group_types = ['raw_datas', 'scan', 'detector', 'move', 'data', 'ch', '']
group_data_types = ['data0D', 'data1D', 'data2D']
data_types = ['data', 'axis', 'live_scan', 'navigation_axis']
data_dimensions = ['0D', '1D', '2D']
scan_types = ['', 'scan1D', 'scan2D']


class H5Saver(QObject):
    status_sig = pyqtSignal(utils.ThreadCommand)
    new_file_sig = pyqtSignal(bool)

    params = [  {'title': 'Save type:', 'name': 'save_type', 'type': 'list', 'values': save_types, 'readonly': True},
                {'title': 'Save 2D datas:', 'name': 'save_2D', 'type': 'bool', 'value': True},
                {'title': 'Save raw datas only:', 'name': 'save_raw_only', 'type': 'bool', 'value': True, 'tooltip':
                        'if True, will not save extracted ROIs used to do live plotting, only raw datas and the scan \
                        result will be saved'},
                {'title': 'Do Save:', 'name': 'do_save', 'type': 'bool', 'default': False, 'value': False},
                {'title': 'N saved:', 'name': 'N_saved', 'type': 'int', 'default': 0, 'value': 0, 'visible': False},
                {'title': 'custom_name?:', 'name': 'custom_name', 'type': 'bool', 'default': False, 'value': False},
                {'title': 'show file content?:', 'name': 'show_file', 'type': 'bool', 'default': False, 'value': False},
                {'title': 'Base path:', 'name': 'base_path', 'type': 'browsepath', 'value': 'C:\Data', 'filetype': False, 'readonly': True,},
                {'title': 'Base name:', 'name': 'base_name', 'type': 'str', 'value': 'Scan', 'readonly': True},
                {'title': 'Current scan:', 'name': 'current_scan_name', 'type': 'str', 'value': '', 'readonly': True},
                {'title': 'Current path:', 'name': 'current_scan_path', 'type': 'text', 'value': 'C:\Data',
                 'readonly': True, 'visible': False},
                {'title': 'h5file:', 'name': 'current_h5_file', 'type': 'text_pb', 'value': '', 'readonly': True},
                {'title': 'Compression options:', 'name': 'compression_options', 'type': 'group', 'children': [
                    {'title': 'Compression library:', 'name': 'h5comp_library', 'type': 'list', 'value': 'zlib',
                     'values': ['zlib', 'lzo', 'bzip2', 'blosc']},
                    {'title': 'Compression level:', 'name': 'h5comp_level', 'type': 'int', 'value': 5, 'min': 0, 'max': 9},
                ]},
                ]

    def __init__(self, h5_file_path=None, h5_file=None, save_type='scan'):
        """
        Initialize the h5Saver object
        Parameters
        ----------
        h5_file_path: (Path) Path object pointing to the h5_file
        h5_file: instance of a h5 file as opened using the pytables module
        save_type: (str) either 'scan', 'detector' or 'custom'
        """
        super(H5Saver, self).__init__()

        if save_type not in save_types:
            raise Exception('Invalid saving type')

        self.h5_file = h5_file
        self.h5_file_path = h5_file_path
        self.h5_file_name = None
        self.logger_array = None

        self.current_group = None
        self.current_scan_group = None
        self.current_scan_name = None
        self.raw_group = None

        self.settings=Parameter.create(title='Saving settings', name='save_settings', type='group', children=self.params)
        self.settings_tree = ParameterTree()
        self.settings_tree.setMinimumHeight(310)
        self.settings_tree.setParameters(self.settings, showTop=False)
        self.settings.child(('current_h5_file')).sigActivated.connect(lambda: self.emit_new_file(True))

        self.settings.child(('save_type')).setValue(save_type)

        self.filters = tables.Filters(
            complevel=self.settings.child('compression_options', 'h5comp_level').value(),
            complib=self.settings.child('compression_options', 'h5comp_library').value())
        # self.settings.child('saving_options', 'save_independent').show(save_type == 'scan')
        # self.settings.child('saving_options', 'do_save').show(not save_type == 'scan')
        # self.settings.child('saving_options', 'current_scan_name').show(save_type == 'scan')


        self.settings.sigTreeStateChanged.connect(self.parameter_tree_changed)  # any changes on the settings will update accordingly the detector

    def emit_new_file(self, status):
        self.new_file_sig.emit(status)

    def init_file(self, update_h5=False, custom_naming=False, addhoc_file_path = None):
        """
        init a new h5 file. Could be a file with a given name (addhoc_file_path) or following the template for scans
        (datasets) or for detectors or a box to set a custom name (custom naming)
        Parameters
        ----------
        update_h5: (bool) create a new h5 file with name specified by other parameters if false try to open an existing
        file or create it if it doesn't exists
        custom_naming: (bool) if True, a selection file dialog opens to set a new file name
        addhoc_file_path: (Path) supplied name for the file

        Returns
        -------
        bool: True if new file has been created, False otherwise
        """
        date = datetime.datetime.now()

        if addhoc_file_path is None:
            if not os.path.isdir(self.settings.child(('base_path')).value()):
                os.mkdir(self.settings.child(('base_path')).value())

            # set the filename and path
            base_name = self.settings.child(('base_name')).value()

            if not custom_naming:
                custom_naming = self.settings.child(('custom_name')).value()

            if not custom_naming:
                scan_type = self.settings.child(('save_type')).value() == 'scan'
                scan_path, current_scan_name, save_path = self.update_file_paths(update_h5)
                self.current_scan_name = current_scan_name
                self.settings.child(('current_scan_name')).setValue(current_scan_name)
                self.settings.child(('current_scan_path')).setValue(str(scan_path))

                if not scan_type:
                    self.h5_file_path = save_path #will remove the dataset part used for DAQ_scan datas
                    self.h5_file_name = base_name+date.strftime('_%Y%m%d_%H_%M_%S.h5')
                else:
                    self.h5_file_path = save_path
                    self.h5_file_name = save_path.name+".h5"
            else:
                self.h5_file_name = utils.select_file(start_path=base_name, save=True, ext='h5')
                self.h5_file_path = self.h5_file_name.parent

        else:
            if isinstance(addhoc_file_path, str):
                addhoc_file_path = Path(addhoc_file_path)
            self.h5_file_path = addhoc_file_path.parent
            self.h5_file_name = addhoc_file_path.name

        fullpathname = str(self.h5_file_path.joinpath(self.h5_file_name))
        self.settings.child(('current_h5_file')).setValue(fullpathname)

        if update_h5:
            self.current_scan_group = None

        scan_group = None
        if self.current_scan_group is not None:
            scan_group = self.current_scan_group._v_name

        if update_h5:
            self.close_file()
            self.h5_file = tables.open_file(fullpathname,'w', title='PyMoDAQ file')
            self.h5_file.root._v_attrs['pymodaq_version'] = get_version()
        else:
            self.close_file()
            self.h5_file = tables.open_file(fullpathname, 'a', title='PyMoDAQ file')

        self.raw_group = self.get_set_group(self.h5_file.root, 'Raw_datas', title='Data from PyMoDAQ modules')
        self.get_set_logger(self.raw_group)
        if scan_group is not None:
            self.current_scan_group = self.get_set_group(self.raw_group, scan_group)
        else:
            self.current_scan_group = self.get_last_scan()

        self.raw_group._v_attrs['type'] = self.settings.child(('save_type')).value()
        self.h5_file.root._v_attrs['file'] = date.strftime(self.h5_file_name)
        if update_h5:
            self.h5_file.root._v_attrs['date'] = date.strftime('%Y%m%d')
            self.h5_file.root._v_attrs['time'] = date.strftime('%H:%M:%S')




        return update_h5

    def update_file_paths(self,update_h5=False):
        """
        """
        try:
            # set the filename and path
            base_path = self.settings.child(('base_path')).value()
            base_name = self.settings.child(('base_name')).value()
            current_scan = self.settings.child(('current_scan_name')).value()
            scan_type = self.settings.child(('save_type')).value() == 'scan'

            if current_scan == '' or update_h5:
                next_scan_index = 0
                update_h5 = True #just started the main program so one should create a new h5
            else:
                next_scan_index = self.get_scan_index()

            scan_path, current_filename, dataset_path=utils.set_current_scan_path(base_path, base_name, update_h5,
                                next_scan_index, create_dataset_folder=scan_type)
            self.settings.child(('current_scan_path')).setValue(str(scan_path))

            return scan_path, current_filename, dataset_path


        except Exception as e:
            print(e)


    def get_last_scan(self):
        groups = [group for group in list(self.raw_group._v_groups) if 'Scan' in group]
        groups.sort()
        if len(groups) != 0:
            scan_group = self.h5_file.get_node(self.raw_group, groups[-1])
        else:
            scan_group = None
        return scan_group


    def get_scan_index(self):
        try:
            if self.current_scan_group is None:
                return 0
            else:

                groups = [group for group in list(self.raw_group._v_groups) if 'Scan' in group]
                groups.sort()
                flag = False
                for child in list(self.h5_file.get_node(self.raw_group, groups[-1])._v_groups):
                    if 'scan' in child:
                        return len(groups)

                return 0

        except Exception as e:
            return 0


    def load_file(self, base_path=None, file_path=None):
        if base_path is None:
            base_path = self.settings.child('base_path').value()
            if not os.path.isdir(base_path):
                base_path = None

        if file_path is None:
            file_path = utils.select_file(base_path, save=False, ext='h5')

        if not isinstance(file_path, Path):
            file_path = Path(file_path)

        if not file_path.suffix == 'h5':
            raise IOError('Invalid file type, should be a h5 file')

        self.init_file(addhoc_file_path=file_path)


    def close_file(self):
        try:
            if self.h5_file is not None:
                self.h5_file.flush()
                if self.h5_file.isopen:
                    self.h5_file.close()
        except Exception as e:
            print(e) #no big deal


    def is_node_in_group(self,where, name):
        """
        Check if a given node with name is in the group defined by where (comparison on lower case strings)
        Parameters
        ----------
        where: (str or node) path or parent node instance
        name: (str) group node name

        Returns
        -------
        bool: True if node exists, False otherwise
        """

        nodes_names = [node._v_name.lower() for node in self.h5_file.list_nodes(where)]
        return name.lower() in nodes_names

    def get_set_logger(self, where):
        logger = 'Logger'
        if not logger in list(self.h5_file.get_node(where)._v_children.keys()):
            # check if logger node exist
            text_atom = tables.atom.ObjectAtom()
            self.logger_array = self.h5_file.create_vlarray(where, logger, atom=text_atom)
            self.logger_array._v_attrs['type'] = 'log'
        else:
            self.logger_array = self.h5_file.get_node(where, name=logger)
        return self.logger_array


    def get_set_group(self, where, name, title=''):
        """
        Retrieve or create (if absent) a node group
        Parameters
        ----------
        where: (str or node) path or parent node instance
        name: (str) group node name
        title: (str) node title

        Returns
        -------
        node: group node
        """
        if not name in list(self.h5_file.get_node(where)._v_children.keys()):
            self.current_group = self.h5_file.create_group(where, name, title)
        else:
            self.current_group = self.h5_file.get_node(where, name)
        return self.current_group

    def add_data_group(self,where, group_data_type, title='', settings_as_xml='', metadata=dict([])):
        if group_data_type not in group_data_types:
            raise Exception('Invalid data group type')
        group = self.add_group(group_data_type, '', where, title, settings_as_xml, metadata)
        return group

    def add_navigation_axis(self, data, parent_group, axis='x_axis', enlargeable=False, title='', metadata=dict([])):
        """
        Create carray for navigation axis within a scan
        Parameters
        ----------
        data: (ndarray) of dimension 1
        parent_group: (str or node) parent node where to save new data
        axis: (str) either x_axis or y_axis
        """
        if axis not in ['x_axis', 'y_axis', 'z_axis']:
            raise Exception('Invalid navigation axis name')
        array = self.add_array(parent_group, 'scan_{:s}'.format(axis), 'navigation_axis', data_shape=data.shape,
                    data_dimension='1D', array_to_save=data, enlargeable=enlargeable, title=title, metadata=metadata)
        return array

    def add_data_live_scan(self,channel_group, data_dict, scan_type='scan1D', title=''):
        shape, dimension, size = utils.get_data_dimension(data_dict['data'], scan_type=scan_type, remove_scan_dimension=True)
        data_array = self.add_array(channel_group, 'Data', 'data', array_type=np.float,
                                    title=title,
                                    data_shape=shape,
                                    data_dimension=dimension, scan_type=scan_type,
                                    array_to_save=data_dict['data'])
        if 'x_axis' in data_dict:
            if not isinstance(data_dict['x_axis'], dict):
                array_to_save = data_dict['x_axis']
                tmp_dict = dict(label='', units='')
            else:
                tmp_dict = copy.deepcopy(data_dict['x_axis'])
                array_to_save = tmp_dict.pop('data')
        if 'x_axis' in data_dict:

            array = self.add_array(channel_group, 'x_axis', 'axis',
                                   array_type=np.float, array_to_save=array_to_save,
                                   enlargeable=False, data_dimension='1D', metadata=tmp_dict)
        if 'y_axis' in data_dict:
            if not isinstance(data_dict['y_axis'], dict):
                array_to_save = data_dict['y_axis']
                tmp_dict = dict(label='', units='')
            else:
                tmp_dict = copy.deepcopy(data_dict['y_axis'])
                array_to_save = tmp_dict.pop('data')
        if 'y_axis' in data_dict:
            array = self.add_array(channel_group, 'y_axis', 'axis',
                                   array_type=np.float, array_to_save=array_to_save,
                                   enlargeable=False, data_dimension='1D', metadata=tmp_dict)
        return data_array

    def add_data(self, channel_group, data_dict, scan_type='scan1D', scan_shape = [], title='', enlargeable=False, init=False, add_scan_dim=False):

        shape, dimension, size = utils.get_data_dimension(data_dict['data'])
        data_array = self.add_array(channel_group, 'Data', 'data', array_type=np.float,
            title=title, data_shape=shape,enlargeable=enlargeable, data_dimension = dimension, scan_type=scan_type,
            scan_shape = scan_shape,
            array_to_save=data_dict['data'],
            init = init, add_scan_dim = add_scan_dim)

        if 'x_axis' in data_dict:
            if not isinstance(data_dict['x_axis'], dict):
                array_to_save = data_dict['x_axis']
                tmp_dict = dict(label='', units='')
            else:
                tmp_dict = copy.deepcopy(data_dict['x_axis'])
                array_to_save = tmp_dict.pop('data')

            array = self.add_array(channel_group, 'x_axis', 'axis',
                   array_type=np.float,array_to_save=array_to_save,
                            enlargeable=False, data_dimension='1D', metadata=tmp_dict)

        if 'y_axis' in data_dict:
            if not isinstance(data_dict['y_axis'], dict):
                array_to_save = data_dict['y_axis']
                tmp_dict = dict(label='', units='')
            else:
                tmp_dict = copy.deepcopy(data_dict['y_axis'])
                array_to_save = tmp_dict.pop('data')

            array = self.add_array(channel_group, 'y_axis', 'axis',
                    array_type=np.float, array_to_save=array_to_save,
                    enlargeable=False, data_dimension='1D', metadata=tmp_dict)
        return data_array


    def add_array(self, where, name, data_type, data_shape=(1,), data_dimension = '0D', scan_type='', scan_shape=[] ,
                  title='', array_to_save=None, array_type = np.float, enlargeable = False, metadata=dict([]),
                  init=False, add_scan_dim=False):

        if data_dimension not in data_dimensions:
            raise Exception('Invalid data dimension')
        if data_type not in data_types:
            raise Exception('Invalid data type')
        if scan_type != '':
            scan_type = utils.uncapitalize(scan_type)
        if scan_type not in scan_types:
            raise Exception('Invalid scan type')
        if enlargeable:
            shape = [0]
            if data_shape != (1,):
                shape.extend(data_shape)
            shape = tuple(shape)
            array = self.h5_file.create_earray(where, utils.capitalize(name), tables.Atom.from_dtype(np.dtype(array_type)), shape=shape,
                                               title=title, filters=self.filters)
            array._v_attrs['shape'] = shape
        else:
            if add_scan_dim:  #means it is an array initialization to zero
                shape = scan_shape[:]
                shape.extend(data_shape)
                if init or array_to_save is None:
                    array_to_save = np.zeros(shape)

            array = self.h5_file.create_carray(where, utils.capitalize(name), obj = array_to_save,
                                               title=title,
                                               filters=self.filters)
            array._v_attrs['shape'] = array_to_save.shape

        array._v_attrs['type'] = data_type
        array._v_attrs['data_dimension'] = data_dimension
        array._v_attrs['scan_type'] = scan_type


        for metadat in metadata:
            array._v_attrs[metadat] = metadata[metadat]
        return array

    def append(self, array, data):
        if not (isinstance(array, tables.vlarray.VLArray) or isinstance(array, tables.earray.EArray)):
            raise Exception('This array cannot be appended')
        if isinstance(data, np.ndarray):
            if data.shape != (1,):
                shape = [1]
                shape.extend(data.shape)
                array.append(data.reshape(shape))
            else:
                array.append(data)
        else:
            array.append(data)
        sh = list(array._v_attrs['shape'])
        sh[0] += 1
        array._v_attrs['shape'] = tuple(sh)

    def add_group(self, group_name, group_type, where, title='', settings_as_xml='', metadata=dict([])):
        """
        Add a node in the h5 file tree of the group type
        Parameters
        ----------
        group_name: (str) a custom name for this group
        group_type: (str) one of the possible values of **group_types**
        where: (str or node) parent node where to create the new group
        settings_as_xml: (str) XML string containing Parameters representation (see custom_Tree)
        metadata: (dict) extra metadata to be saved with this new group node

        Returns
        -------
        (node): newly created group node
        """

        if group_type not in group_types:
            raise Exception('Invalid group type')

        try:
            node = self.h5_file.get_node(where, utils.capitalize(group_name))
        except tables.NoSuchNodeError as e:
            node = None

        if node is None:
            node = self.get_set_group(where, utils.capitalize(group_name), title)
            node._v_attrs['settings'] = settings_as_xml
            node._v_attrs['type'] = group_type.lower()
            for metadat in metadata:
                node._v_attrs[metadat] = metadata[metadat]

        return node

    def add_incremental_group(self, group_type, where, title='', settings_as_xml='', metadata=dict([])):
        """
        Add a node in the h5 file tree of the group type with an increment in the given name
        Parameters
        ----------
        group_type: (str) one of the possible values of **group_types**
        where: (str or node) parent node where to create the new group
        settings_as_xml: (str) XML string containing Parameters representation (see custom_Tree)
        metadata: (dict) extra metadata to be saved with this new group node

        Returns
        -------
        (node): newly created group node
        """
        if group_type not in group_types:
            raise Exception('Invalid group type')
        nodes = list(self.h5_file.get_node(where)._v_children.keys())
        nodes_tmp = []
        for node in nodes:
            if utils.capitalize(group_type) in node:
                nodes_tmp.append(node)
        nodes_tmp.sort()
        if len(nodes_tmp) ==0:
            ind_group = -1
        else:
            ind_group = int(nodes_tmp[-1][-3:])
        group = self.get_set_group(where, utils.capitalize(group_type)+'{:03d}'.format(ind_group + 1), title)
        group._v_attrs['settings'] = settings_as_xml
        if group_type.lower() != 'ch':
            group._v_attrs['type'] = group_type.lower()
        else:
            group._v_attrs['type'] = ''
        for metadat in metadata:
            group._v_attrs[metadat] = metadata[metadat]
        return group


    def add_det_group(self, where, title='', settings_as_xml='', metadata=dict([])):
        """
        Add a new group of type detector
        See Also
        -------
        add_incremental_group
        """
        group = self.add_incremental_group('detector', where, title, settings_as_xml, metadata)
        return group

    def add_CH_group(self, where, title='', settings_as_xml='', metadata=dict([])):
        """
        Add a new group of type channel
        See Also
        -------
        add_incremental_group
        """
        group = self.add_incremental_group('ch', where, title, settings_as_xml, metadata)
        return group

    def add_live_scan_group(self, where, dimensionality, title='', settings_as_xml='', metadata=dict([])):
        """
        Add a new group of type live scan
        See Also
        -------
        add_incremental_group
        """
        group = self.add_group('Live_scan_{:s}'.format(dimensionality), '', where, title=title,
                               settings_as_xml=settings_as_xml, metadata=metadata)
        return group

    def add_scan_group(self, title='', settings_as_xml='', metadata=dict([])):
        """
        Add a new group of type scan
        See Also
        -------
        add_incremental_group
        """

        if self.current_scan_group is not None:
            if list(self.current_scan_group._v_children) == []:
                new_scan = False
            else:
                new_scan = True
        else:
            new_scan = True
        if new_scan:
            self.current_scan_group = self.add_incremental_group('scan', self.raw_group, title, settings_as_xml, metadata)
            self.current_scan_group._v_attrs['description'] = ''
            self.settings.child(('current_scan_name')).setValue(self.current_scan_group._v_name)


        return self.current_scan_group

    def add_move_group(self, where, title='', settings_as_xml='', metadata=dict([])):
        """
        Add a new group of type move
        See Also
        -------
        add_incremental_group
        """
        group = self.add_incremental_group('move', where, title, settings_as_xml, metadata)
        return group


    def parameter_tree_changed(self,param,changes):
        for param, change, data in changes:
            path = self.settings.childPath(param)

            if change == 'childAdded':
                pass

            elif change == 'value':
                if param.name() == 'show_file':
                    param.setValue(False)
                    self.show_file_content()


                elif param.name() == 'base_path':
                    try:
                        if not os.path.isdir(param.value()):
                            os.mkdir(param.value())
                    except:
                        self.update_status("The base path couldn't be set, please check your options")

                elif param.name() in custom_tree.iter_children(self.settings.child(('compression_options')), []):
                    self.filters = tables.Filters(
                        complevel=self.settings.child('compression_options', 'h5comp_level').value(),
                        complib=self.settings.child('compression_options', 'h5comp_library').value())

            elif change == 'parent':
                pass

    def update_status(self,status):
        self.status_sig.emit(utils.ThreadCommand("Update_Status", [status, 'log']))

    def show_file_content(self):
        form = QtWidgets.QWidget()
        if not self.h5_file.isopen:
            if self.h5_file_path.exists():
                self.analysis_prog = H5Browser(form,h5file=self.h5_file_path)
            else:
                raise FileExistsError('no File presents')
        else:
            self.analysis_prog = H5Browser(form,h5file=self.h5_file)
        form.show()



if __name__ == '__main__':
    import sys
    app = QtWidgets.QApplication(sys.argv)
    prog = H5Saver()
    prog.settings_tree.show()

    prog.init_file(True)

    #prog.add_scan_group()

    sys.exit(app.exec_())